@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2018-2020 Tada AB and other contributors, as listed below.
+ * Copyright (c) 2018-2021 Tada AB and other contributors, as listed below.
  *
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the The BSD 3-Clause License
@@ -50,6 +50,7 @@ import javax.xml.transform.Transformer;
 import javax.xml.transform.TransformerFactory;
 
 import javax.xml.transform.TransformerException;
+import javax.xml.transform.TransformerConfigurationException;
 
 import javax.xml.transform.stream.StreamResult;
 import javax.xml.transform.stream.StreamSource;
@@ -67,14 +68,14 @@ import org.postgresql.pljava.Adjusting;
 import org.postgresql.pljava.annotation.Function;
 import org.postgresql.pljava.annotation.MappedUDT;
 import org.postgresql.pljava.annotation.SQLAction;
-import org.postgresql.pljava.annotation.SQLActions;
 import org.postgresql.pljava.annotation.SQLType;
 
 import static org.postgresql.pljava.example.LoggerTest.logMessage;
 
 /* Imports needed just for the SAX flavor of "low-level XML echo" below */
+import javax.xml.parsers.ParserConfigurationException;
+import javax.xml.parsers.SAXParserFactory;
 import org.xml.sax.XMLReader;
-import org.xml.sax.helpers.XMLReaderFactory;
 import org.xml.sax.ContentHandler;
 import org.xml.sax.DTDHandler;
 import org.xml.sax.ext.LexicalHandler;
@@ -104,114 +105,180 @@ import org.w3c.dom.bootstrap.DOMImplementationRegistry;
  * Everything mentioning the type XML here needs a conditional implementor tag
  * in case of being loaded into a PostgreSQL instance built without that type.
  */
-@SQLActions({
-	@SQLAction(provides="postgresql_xml", install=
-		"SELECT CASE (SELECT 1 FROM pg_type WHERE typname = 'xml') WHEN 1" +
-		" THEN set_config('pljava.implementors', 'postgresql_xml,' || " +
-		" current_setting('pljava.implementors'), true) " +
-		"END"
-	),
+@SQLAction(provides="postgresql_xml", install=
+	"SELECT CASE (SELECT 1 FROM pg_type WHERE typname = 'xml') WHEN 1" +
+	" THEN set_config('pljava.implementors', 'postgresql_xml,' || " +
+	" current_setting('pljava.implementors'), true) " +
+	"END"
+)
 
-	@SQLAction(implementor="postgresql_ge_80400",
-		provides="postgresql_xml_ge84",
-		install=
-		"SELECT CASE (SELECT 1 FROM pg_type WHERE typname = 'xml') WHEN 1" +
-		" THEN set_config('pljava.implementors', 'postgresql_xml_ge84,' || " +
-		" current_setting('pljava.implementors'), true) " +
-		"END"
-	),
+@SQLAction(implementor="postgresql_ge_80400",
+	provides="postgresql_xml_ge84",
+	install=
+	"SELECT CASE (SELECT 1 FROM pg_type WHERE typname = 'xml') WHEN 1" +
+	" THEN set_config('pljava.implementors', 'postgresql_xml_ge84,' || " +
+	" current_setting('pljava.implementors'), true) " +
+	"END"
+)
+
+@SQLAction(implementor="postgresql_xml_ge84", requires="echoXMLParameter",
+	install=
+	"WITH" +
+	" s(how) AS (SELECT generate_series(1, 7))," +
+	" t(x) AS (" +
+	"  SELECT table_to_xml('pg_catalog.pg_operator', true, false, '')" +
+	" )," +
+	" r(howin, howout, isdoc) AS (" +
+	"  SELECT" +
+	"   i.how, o.how," +
+	"   javatest.echoxmlparameter(x, i.how, o.how) IS DOCUMENT" +
+	"  FROM" +
+	"   t, s AS i, s AS o" +
+	"  WHERE" +
+	"   NOT (i.how = 6 and o.how = 7)" + // 6->7 unreliable in some JREs
+	" ) " +
+	"SELECT" +
+	" CASE WHEN every(isdoc)" +
+	"  THEN javatest.logmessage('INFO', 'SQLXML echos succeeded')" +
+	"  ELSE javatest.logmessage('WARNING', 'SQLXML echos had problems')" +
+	" END " +
+	"FROM" +
+	" r"
+)
+
+@SQLAction(implementor="postgresql_xml_ge84", requires="proxiedXMLEcho",
+	install=
+	"WITH" +
+	" s(how) AS (SELECT unnest('{1,2,4,5,6,7}'::int[]))," +
+	" t(x) AS (" +
+	"  SELECT table_to_xml('pg_catalog.pg_operator', true, false, '')" +
+	" )," +
+	" r(how, isdoc) AS (" +
+	"  SELECT" +
+	"	how," +
+	"	javatest.proxiedxmlecho(x, how) IS DOCUMENT" +
+	"  FROM" +
+	"	t, s" +
+	" )" +
+	"SELECT" +
+	" CASE WHEN every(isdoc)" +
+	"  THEN javatest.logmessage('INFO', 'proxied SQLXML echos succeeded')" +
+	"  ELSE javatest.logmessage('WARNING'," +
+	"       'proxied SQLXML echos had problems')" +
+	" END " +
+	"FROM" +
+	" r"
+)
+
+@SQLAction(implementor="postgresql_xml_ge84", requires="lowLevelXMLEcho",
+	install={
+	"SELECT" +
+	" preparexmlschema('schematest', $$" +
+	"<xs:schema" +
+	" xmlns:xs='http://www.w3.org/2001/XMLSchema'" +
+	" targetNamespace='urn:testme'" +
+	" elementFormDefault='qualified'>" +
+	" <xs:element name='row'>" +
+	"  <xs:complexType>" +
+	"   <xs:sequence>" +
+	"    <xs:element name='textcol' type='xs:string' nillable='true'/>" +
+	"    <xs:element name='intcol' type='xs:integer' nillable='true'/>" +
+	"   </xs:sequence>" +
+	"  </xs:complexType>" +
+	" </xs:element>" +
+	"</xs:schema>" +
+	"$$, 'http://www.w3.org/2001/XMLSchema', 5)",
+
+	"WITH" +
+	" s(how) AS (SELECT unnest('{4,5,7}'::int[]))," +
+	" r(isdoc) AS (" +
+	" SELECT" +
+	"  javatest.lowlevelxmlecho(" +
+	"   query_to_xml(" +
+	"    'SELECT ''hi'' AS textcol, 1 AS intcol', true, true, 'urn:testme'"+
+	"   ), how, params) IS DOCUMENT" +
+	" FROM" +
+	"  s," +
+	"  (SELECT 'schematest' AS schema) AS params" +
+	" )" +
+	"SELECT" +
+	" CASE WHEN every(isdoc)" +
+	"  THEN javatest.logmessage('INFO', 'XML Schema tests succeeded')" +
+	"  ELSE javatest.logmessage('WARNING'," +
+	"       'XML Schema tests had problems')" +
+	" END " +
+	"FROM" +
+	" r"
+	}
+)
+
+@SQLAction(implementor="postgresql_xml",
+		   requires={"prepareXMLTransform", "transformXML"},
+	install={
+		"REVOKE EXECUTE ON FUNCTION javatest.prepareXMLTransformWithJava" +
+		" (pg_catalog.varchar, pg_catalog.xml, integer, boolean, boolean," +
+		"  pg_catalog.RECORD)" +
+		" FROM PUBLIC",
 
-	@SQLAction(implementor="postgresql_xml_ge84", requires="echoXMLParameter",
-		install=
-		"WITH" +
-		" s(how) AS (SELECT generate_series(1, 7))," +
-		" t(x) AS (" +
-		"  SELECT table_to_xml('pg_catalog.pg_operator', true, false, '')" +
-		" )," +
-		" r(howin, howout, isdoc) AS (" +
-		"  SELECT" +
-		"   i.how, o.how," +
-		"   javatest.echoxmlparameter(x, i.how, o.how) IS DOCUMENT" +
-		"  FROM" +
-		"   t, s AS i, s AS o" +
-		"  WHERE" +
-		"   NOT (i.how = 6 and o.how = 7)" + // 6->7 unreliable in some JREs
-		" ) " +
 		"SELECT" +
-		" CASE WHEN every(isdoc)" +
-		"  THEN javatest.logmessage('INFO', 'SQLXML echos succeeded')" +
-		"  ELSE javatest.logmessage('WARNING', 'SQLXML echos had problems')" +
-		" END " +
-		"FROM" +
-		" r"
-	),
+		" javatest.prepareXMLTransform('distinctElementNames'," +
+		"'<xsl:transform version=''1.0''" +
+		" xmlns:xsl=''http://www.w3.org/1999/XSL/Transform''" +
+		" xmlns:exsl=''http://exslt.org/common''" +
+		" xmlns:set=''http://exslt.org/sets''" +
+		" extension-element-prefixes=''exsl set''" +
+		">" +
+		" <xsl:output method=''xml'' indent=''no''/>" +
+		" <xsl:template match=''/''>" +
+		"  <xsl:variable name=''enames''>" +
+		"   <xsl:for-each select=''//*''>" +
+		"    <ename><xsl:value-of select=''local-name()''/></ename>" +
+		"   </xsl:for-each>" +
+		"  </xsl:variable>" +
+		"  <xsl:for-each" +
+		"   select=''set:distinct(exsl:node-set($enames)/ename)''>" +
+		"   <xsl:sort select=''string()''/>" +
+		"   <den><xsl:value-of select=''.''/></den>" +
+		"  </xsl:for-each>" +
+		" </xsl:template>" +
+		"</xsl:transform>', how => 5, enableExtensionFunctions => true)",
 
-	@SQLAction(implementor="postgresql_xml_ge84", requires="proxiedXMLEcho",
-		install=
-		"WITH" +
-		" s(how) AS (SELECT unnest('{1,2,4,5,6,7}'::int[]))," +
-		" t(x) AS (" +
-		"  SELECT table_to_xml('pg_catalog.pg_operator', true, false, '')" +
-		" )," +
-		" r(how, isdoc) AS (" +
-		"  SELECT" +
-		"	how," +
-		"	javatest.proxiedxmlecho(x, how) IS DOCUMENT" +
-		"  FROM" +
-		"	t, s" +
-		" )" +
 		"SELECT" +
-		" CASE WHEN every(isdoc)" +
-		"  THEN javatest.logmessage('INFO', 'proxied SQLXML echos succeeded')" +
-		"  ELSE javatest.logmessage('WARNING'," +
-		"       'proxied SQLXML echos had problems')" +
-		" END " +
-		"FROM" +
-		" r"
-	),
+		" javatest.prepareXMLTransformWithJava('getPLJavaVersion'," +
+		"'<xsl:transform version=''1.0''" +
+		" xmlns:xsl=''http://www.w3.org/1999/XSL/Transform''" +
+		" xmlns:java=''http://xml.apache.org/xalan/java''" +
+		" exclude-result-prefixes=''java''" +
+		">" +
+		" <xsl:template match=''/''>" +
+		"  <xsl:value-of" +
+		"   select=''java:java.lang.System.getProperty(" +
+		"    \"org.postgresql.pljava.version\")''" +
+		"  />" +
+		" </xsl:template>" +
+		"</xsl:transform>', enableExtensionFunctions => true)",
 
-	@SQLAction(implementor="postgresql_xml_ge84", requires="lowLevelXMLEcho",
-		install={
 		"SELECT" +
-		" preparexmlschema('schematest', $$" +
-		"<xs:schema" +
-		" xmlns:xs='http://www.w3.org/2001/XMLSchema'" +
-		" targetNamespace='urn:testme'" +
-		" elementFormDefault='qualified'>" +
-		" <xs:element name='row'>" +
-		"  <xs:complexType>" +
-		"   <xs:sequence>" +
-		"    <xs:element name='textcol' type='xs:string' nillable='true'/>" +
-		"    <xs:element name='intcol' type='xs:integer' nillable='true'/>" +
-		"   </xs:sequence>" +
-		"  </xs:complexType>" +
-		" </xs:element>" +
-		"</xs:schema>" +
-		"$$, 'http://www.w3.org/2001/XMLSchema', 5)",
+		" CASE WHEN" +
+		"  javatest.transformXML('distinctElementNames'," +
+		"   '<a><c/><e/><b/><b/><d/></a>', 5, 5)::text" +
+		"  =" +
+		"   '<den>a</den><den>b</den><den>c</den><den>d</den><den>e</den>'"+
+		"  THEN javatest.logmessage('INFO', 'XSLT 1.0 test succeeded')" +
+		"  ELSE javatest.logmessage('WARNING', 'XSLT 1.0 test failed')" +
+		" END",
 
-		"WITH" +
-		" s(how) AS (SELECT unnest('{4,5,7}'::int[]))," +
-		" r(isdoc) AS (" +
-		" SELECT" +
-		"  javatest.lowlevelxmlecho(" +
-		"   query_to_xml(" +
-		"    'SELECT ''hi'' AS textcol, 1 AS intcol', true, true, 'urn:testme'"+
-		"   ), how, params) IS DOCUMENT" +
-		" FROM" +
-		"  s," +
-		"  (SELECT 'schematest' AS schema) AS params" +
-		" )" +
 		"SELECT" +
-		" CASE WHEN every(isdoc)" +
-		"  THEN javatest.logmessage('INFO', 'XML Schema tests succeeded')" +
-		"  ELSE javatest.logmessage('WARNING'," +
-		"       'XML Schema tests had problems')" +
-		" END " +
-		"FROM" +
-		" r"
-		}
-	)
-})
+		" CASE WHEN" +
+		"  javatest.transformXML('getPLJavaVersion', '')::text" +
+		"  OPERATOR(pg_catalog.=) extversion" +
+		"  THEN javatest.logmessage('INFO', 'XSLT 1.0 with Java succeeded')" +
+		"  ELSE javatest.logmessage('WARNING', 'XSLT 1.0 with Java failed')" +
+		" END" +
+		" FROM pg_catalog.pg_extension" +
+		" WHERE extname = 'pljava'"
+	}
+)
 @MappedUDT(schema="javatest", name="onexml", structure="c1 xml",
 		   implementor="postgresql_xml",
            comment="A composite type mapped by the PassXML example class")
@@ -219,11 +286,11 @@ public class PassXML implements SQLData
 {
 	static SQLXML s_sx;
 
-	static TransformerFactory s_tf = TransformerFactory.newInstance();
+	static TransformerFactory s_tf = TransformerFactory.newDefaultInstance();
 
-	static Map<String,Templates> s_tpls = new HashMap<String,Templates>();
+	static Map<String,Templates> s_tpls = new HashMap<>();
 
-	static Map<String,Schema> s_schemas = new HashMap<String,Schema>();
+	static Map<String,Schema> s_schemas = new HashMap<>();
 
 	@Function(schema="javatest", implementor="postgresql_xml")
 	public static String inXMLoutString(SQLXML in) throws SQLException
@@ -359,54 +426,193 @@ public class PassXML implements SQLData
 	 * Each value of {@code how}, 1-7, selects a different way of presenting
 	 * the {@code SQLXML} object to the XSL processor.
 	 *<p>
-	 * Preparing a transform with
-	 * {@link TransformerFactory#newTemplates newTemplates()} seems to require
-	 * {@link Function.Trust#UNSANDBOXED Trust.UNSANDBOXED}, at least for the
-	 * XSLTC transform compiler in newer JREs.
+	 * Passing {@code true} for {@code enableExtensionFunctions} allows the
+	 * transform to use extensions that the Java XSLT implementation supports,
+	 * such as functions from EXSLT. Those are disabled by default.
 	 *<p>
-	 * If you wish this <strong>unsandboxed</strong> function to be installed,
-	 * set the PostgreSQL variable {@code pljava.implementors} to a list with
-	 * {@code pg_xml_unsandboxed} as an added entry, before installing the
-	 * examples jar.
+	 * Passing {@code false} for {@code builtin} will allow a
+	 * {@code TransformerFactory} other than Java's built-in one to be found
+	 * using the usual search order and the context class loader (normally
+	 * the PL/Java class path for the schema where this function is declared).
+	 * The default of {@code true} ensures that the built-in Java XSLT 1.0
+	 * implementation is used. A transformer implementation other than Xalan
+	 * may not recognize the feature controlled by
+	 * {@code enableExtensionFunctions}, so failure to configure that feature
+	 * will be logged as a warning if {@code builtin} is {@code false}, instead
+	 * of thrown as an exception.
+	 *<p>
+	 * Out of the box, Java's transformers only support XSLT 1.0. See the S9
+	 * example for more capabilities (at the cost of downloading the Saxon jar).
 	 */
-	@Function(schema="javatest", trust=Function.Trust.UNSANDBOXED,
-			  implementor="pg_xml_unsandboxed")
-	public static void prepareXMLTransform(String name, SQLXML source, int how)
+	@Function(schema="javatest", implementor="postgresql_xml",
+			  provides="prepareXMLTransform")
+	public static void prepareXMLTransform(String name, SQLXML source,
+		@SQLType(defaultValue="0") int how,
+		@SQLType(defaultValue="false") boolean enableExtensionFunctions,
+		@SQLType(defaultValue="true") boolean builtin,
+		@SQLType(defaultValue={}) ResultSet adjust)
 	throws SQLException
 	{
+		prepareXMLTransform(
+			name, source, how, enableExtensionFunctions, adjust, builtin,
+			/* withJava */ false);
+	}
+
+	/**
+	 * Precompile an XSL transform {@code source} and save it (for the
+	 * current session) as {@code name}, where the transform may call Java
+	 * methods.
+	 *<p>
+	 * Otherwise identical to {@code prepareXMLTransform}, this version sets the
+	 * {@code TransformerFactory}'s {@code extensionClassLoader} (to the context
+	 * class loader, normally the PL/Java class path for the schema where this
+	 * function is declared), so the transform will be able to use
+	 * xalan's Java call syntax to call any public Java methods that would be
+	 * accessible to this class. (That can make a big difference in usefulness
+	 * for the otherwise rather limited XSLT 1.0.)
+	 *<p>
+	 * As with {@code enableExtensionFunctions}, failure by the transformer
+	 * implementation to recognize or allow the {@code extensionClassLoader}
+	 * property will be logged as a warning if {@code builtin} is {@code false},
+	 * rather than thrown as an exception.
+	 *<p>
+	 * This example function will be installed with {@code EXECUTE} permission
+	 * revoked from {@code PUBLIC}, as it essentially confers the ability to
+	 * create arbitrary new Java functions, so should only be granted to roles
+	 * you would be willing to grant {@code USAGE ON LANGUAGE java}.
+	 *<p>
+	 * Because this function only prepares the transform, and
+	 * {@link #transformXML transformXML} applies it, there is some division of
+	 * labor in determining what limits apply to its behavior. The use of this
+	 * method instead of {@code prepareXMLTransform} determines whether the
+	 * transform is allowed to see external Java methods at all; it will be
+	 * the policy permissions granted to {@code transformXML} that control what
+	 * those methods can do when the transform is applied. For now, that method
+	 * is defined in the trusted/sandboxed {@code java} language, so this
+	 * function could reasonably be granted to any role with {@code USAGE} on
+	 * {@code java}. If, by contrast, {@code transformXML} were declared in the
+	 * 'untrusted' {@code javaU}, it would be prudent to allow only superusers
+	 * access to this function, just as only they can {@code CREATE FUNCTION} in
+	 * an untrusted language.
+	 */
+	@Function(schema="javatest", implementor="postgresql_xml",
+			  provides="prepareXMLTransform")
+	public static void prepareXMLTransformWithJava(String name, SQLXML source,
+		@SQLType(defaultValue="0") int how,
+		@SQLType(defaultValue="false") boolean enableExtensionFunctions,
+		@SQLType(defaultValue="true") boolean builtin,
+		@SQLType(defaultValue={}) ResultSet adjust)
+	throws SQLException
+	{
+		prepareXMLTransform(
+			name, source, how, enableExtensionFunctions, adjust, builtin,
+			/* withJava */ true);
+	}
+
+	private static void prepareXMLTransform(String name, SQLXML source, int how,
+		boolean enableExtensionFunctions, ResultSet adjust, boolean builtin,
+		boolean withJava)
+	throws SQLException
+	{
+		TransformerFactory tf =
+			builtin
+			? TransformerFactory.newDefaultInstance()
+			: TransformerFactory.newInstance();
+		String exf =
+		  "http://www.oracle.com/xml/jaxp/properties/enableExtensionFunctions";
+		String ecl = "jdk.xml.transform.extensionClassLoader";
+		Source src = sxToSource(source, how, adjust);
 		try
 		{
-			s_tpls.put(name, s_tf.newTemplates(sxToSource(source, how)));
+			try
+			{
+				tf.setFeature(exf, enableExtensionFunctions);
+			}
+			catch ( TransformerConfigurationException e )
+			{
+				logMessage("WARNING",
+					"non-builtin transformer: ignoring " + e.getMessage());
+			}
+
+			if ( withJava )
+			{
+				try
+				{
+					tf.setAttribute(ecl,
+						Thread.currentThread().getContextClassLoader());
+				}
+				catch ( IllegalArgumentException e )
+				{
+					logMessage("WARNING",
+					"non-builtin transformer: ignoring " + e.getMessage());
+				}
+			}
+
+			s_tpls.put(name, tf.newTemplates(src));
 		}
 		catch ( TransformerException te )
 		{
-			throw new SQLException("XML transformation failed", te);
+			throw new SQLException(
+				"Preparing XML transformation: " + te.getMessage(), te);
 		}
 	}
 
 	/**
 	 * Transform some XML according to a named transform prepared with
 	 * {@code prepareXMLTransform}.
+	 *<p>
+	 * Pass null for {@code transformName} to get a plain identity transform
+	 * (not such an interesting thing to do, unless you also specify indenting).
 	 */
-	@Function(schema="javatest", implementor="postgresql_xml")
+	@Function(schema="javatest", implementor="postgresql_xml",
+			  provides="transformXML")
 	public static SQLXML transformXML(
-		String transformName, SQLXML source, int howin, int howout)
+		String transformName, SQLXML source,
+		@SQLType(defaultValue="0") int howin,
+		@SQLType(defaultValue="0") int howout,
+		@SQLType(defaultValue={}) ResultSet adjust,
+		@SQLType(optional=true) Boolean indent,
+		@SQLType(optional=true) Integer indentWidth)
 	throws SQLException
 	{
-		Templates tpl = s_tpls.get(transformName);
-		Source src = sxToSource(source, howin);
+		Templates tpl = null == transformName? null: s_tpls.get(transformName);
+		Source src = sxToSource(source, howin, adjust);
+
+		if ( Boolean.TRUE.equals(indent)  &&  0 == howout )
+			howout = 4; // transformer only indents if writing a StreamResult
+
 		Connection c = DriverManager.getConnection("jdbc:default:connection");
 		SQLXML result = c.createSQLXML();
-		Result rlt = sxToResult(result, howout);
+		Result rlt = sxToResult(result, howout, adjust);
 
 		try
 		{
-			Transformer t = tpl.newTransformer();
+			Transformer t =
+				null == tpl ? s_tf.newTransformer() : tpl.newTransformer();
+			/*
+			 * For the non-SAX/StAX/DOM flavors of output, you're responsible
+			 * for setting the Transformer to use the server encoding.
+			 */
+			if ( rlt instanceof StreamResult )
+				t.setOutputProperty(ENCODING,
+					System.getProperty("org.postgresql.server.encoding"));
+			else if ( Boolean.TRUE.equals(indent) )
+				logMessage("WARNING",
+					"indent requested, but howout specifies a non-stream " +
+					"Result type; no indenting will happen");
+
+			if ( null != indent )
+				t.setOutputProperty("indent", indent ? "yes" : "no");
+			if ( null != indentWidth )
+				t.setOutputProperty(
+					"{http://xml.apache.org/xalan}indent-amount",
+					"" + indentWidth);
+
 			t.transform(src, rlt);
 		}
 		catch ( TransformerException te )
 		{
-			throw new SQLException("XML transformation failed", te);
+			throw new SQLException("Transforming XML: " + te.getMessage(), te);
 		}
 
 		return ensureClosed(rlt, result, howout);
@@ -688,9 +894,9 @@ public class PassXML implements SQLData
 				XMLReader xr  = sxs.getXMLReader();
 				if ( null == xr )
 				{
-					xr = XMLReaderFactory.createXMLReader();
-					xr.setFeature("http://xml.org/sax/features/namespaces",
-									true);
+					SAXParserFactory spf = SAXParserFactory.newInstance();
+					spf.setNamespaceAware(true);
+					xr = spf.newSAXParser().getXMLReader();
 					/*
 					 * Important: before copying this example code for another
 					 * use, consider whether the input XML might be untrusted.
@@ -778,15 +984,11 @@ public class PassXML implements SQLData
 			throw new SQLException(
 				"IOException in lowLevelXMLEcho", "58030", e);
 		}
-		catch ( SAXException e )
+		catch (
+			ParserConfigurationException | SAXException | XMLStreamException e )
 		{
 			throw new SQLException(
-				"SAXException in lowLevelXMLEcho", "22000", e);
-		}
-		catch ( XMLStreamException e )
-		{
-			throw new SQLException(
-				"XMLStreamException in lowLevelXMLEcho", "22000", e);
+				"XML exception in lowLevelXMLEcho", "22000", e);
 		}
 		return rx;
 	}
@@ -1006,6 +1208,17 @@ public class PassXML implements SQLData
 		}
 	}
 
+
+	/**
+	 * Return some instance of {@code Source} for reading an {@code SQLXML}
+	 * object, depending on the parameter {@code how}.
+	 *<p>
+	 * Note that this method always returns a {@code Source}, even for cases
+	 * 1 and 2 (obtaining readable streams directly from the {@code SQLXML}
+	 * object; this method wraps them in {@code Source}), and case 3
+	 * ({@code getString}; this method creates a {@code StringReader} and
+	 * returns it wrapped in a {@code Source}.
+	 */
 	private static Source sxToSource(SQLXML sx, int how) throws SQLException
 	{
 		switch ( how )
@@ -1052,6 +1265,70 @@ public class PassXML implements SQLData
 		}
 	}
 
+	/**
+	 * Return some instance of {@code Source} for reading an {@code SQLXML}
+	 * object, depending on the parameter {@code how}, applying any adjustments
+	 * in {@code adjust}.
+	 *<p>
+	 * Allows {@code how} to be zero, meaning to let the implementation choose
+	 * what kind of {@code Source} to present. Otherwise identical to the other
+	 * {@code sxToSource}.
+	 */
+	private static Source sxToSource(SQLXML sx, int how, ResultSet adjust)
+	throws SQLException
+	{
+		Source s;
+		switch ( how )
+		{
+			case  0: s = sx.getSource(Adjusting.XML.Source.class); break;
+			case  1:
+			case  2:
+			case  3:
+			case  4:
+				return sxToSource(sx, how); // no adjustments on a StreamSource
+			case  5: s = sx.getSource(Adjusting.XML.SAXSource.class); break;
+			case  6: s = sx.getSource(Adjusting.XML.StAXSource.class); break;
+			case  7: s = sx.getSource(Adjusting.XML.DOMSource.class); break;
+			default: throw new SQLDataException("how should be 0-7", "22003");
+		}
+
+		if ( s instanceof Adjusting.XML.Source )
+			return applyAdjustments(adjust, (Adjusting.XML.Source<?>)s).get();
+		return s;
+	}
+
+	/**
+	 * Return some instance of {@code Result} for writing an {@code SQLXML}
+	 * object, depending on the parameter {@code how} applying any adjustments
+	 * in {@code adjust}.
+	 *<p>
+	 * Allows {@code how} to be zero, meaning to let the implementation choose
+	 * what kind of {@code Result} to present. Otherwise identical to the other
+	 * {@code sxToResult}.
+	 */
+	private static Result sxToResult(SQLXML sx, int how, ResultSet adjust)
+	throws SQLException
+	{
+		Result r;
+		switch ( how )
+		{
+			case  1: // you might wish you could adjust a raw BinaryStream
+			case  2: // or CharacterStream
+			case  3: // or String, but you can't. Ask for a StreamResult.
+			case  5: // SAXResult needs no adjustment
+			case  6: // StAXResult needs no adjustment
+			case  7: // DOMResult needs no adjustment
+				return sxToResult(sx, how);
+			case  4: r = sx.setResult(Adjusting.XML.StreamResult.class); break;
+			case  0: r = sx.setResult(Adjusting.XML.Result.class); break;
+			default: throw new SQLDataException("how should be 0-7", "22003");
+		}
+
+		if ( r instanceof Adjusting.XML.Result )
+			return applyAdjustments(adjust, (Adjusting.XML.Result<?>)r).get();
+		return r;
+	}
+
 	/**
 	 * Ensure the closing of whatever method was used to add content to
 	 * an {@code SQLXML} object.
@@ -1307,6 +1584,7 @@ public class PassXML implements SQLData
 		}
 
 		@Override
+		@SuppressWarnings("unchecked") // all the fun's when sourceClass is null
 		public <T extends Source> T getSource(Class<T> sourceClass)
 		throws SQLException
 		{
@@ -1419,6 +1697,7 @@ public class PassXML implements SQLData
 		}
 
 		@Override
+		@SuppressWarnings("unchecked") // sourceClass==StreamSource is verified
 		public <T extends Source> T getSource(Class<T> sourceClass)
 		throws SQLException
 		{
