@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004-2020 Tada AB and other contributors, as listed below.
+ * Copyright (c) 2004-2021 Tada AB and other contributors, as listed below.
  *
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the The BSD 3-Clause License
@@ -46,6 +46,7 @@
 #include <unistd.h>
 
 #include "org_postgresql_pljava_internal_Backend.h"
+#include "org_postgresql_pljava_internal_Backend_EarlyNatives.h"
 #include "pljava/DualState.h"
 #include "pljava/Invocation.h"
 #include "pljava/InstallHelper.h"
@@ -107,27 +108,23 @@ MemoryContext JavaMemoryContext;
 
 static JavaVM* s_javaVM = 0;
 static jclass  s_Backend_class;
-static jmethodID s_setTrusted;
+static bool    s_startingVM;
 
 /*
  * GUC states
  */
 static char* libjvmlocation;
 static char* vmoptions;
-static char* classpath;
+static char* modulepath;
 static char* implementors;
+static char* policy_urls;
 static int   statementCacheSize;
 static bool  pljavaDebug;
 static bool  pljavaReleaseLingeringSavepoints;
 static bool  pljavaEnabled;
 
-#if PG_VERSION_NUM >= 80400
 static int   java_thread_pg_entry;
-#else
-static char* java_thread_pg_entry;
-#endif
 
-static bool  s_currentTrust;
 static int   s_javaLogLevel;
 
 #if PG_VERSION_NUM < 100000
@@ -135,6 +132,8 @@ bool integerDateTimes = false;
 static void checkIntTimeType(void);
 #endif
 
+static char s_path_var_sep;
+
 extern void Invocation_initialize(void);
 extern void Exception_initialize(void);
 extern void Exception_initialize2(void);
@@ -163,9 +162,11 @@ static void JVMOptList_init(JVMOptList*);
 static void JVMOptList_delete(JVMOptList*);
 static void JVMOptList_add(JVMOptList*, const char*, void*, bool);
 static void JVMOptList_addVisualVMName(JVMOptList*);
+static void JVMOptList_addModuleMain(JVMOptList*);
 static void addUserJVMOptions(JVMOptList*);
-static char* getClassPath(const char*);
-static jint JNICALL my_vfprintf(FILE*, const char*, va_list);
+static char* getModulePath(const char*);
+static jint JNICALL my_vfprintf(FILE*, const char*, va_list)
+	pg_attribute_printf(2, 0);
 static void _destroyJavaVM(int, Datum);
 static void initPLJavaClasses(void);
 static void initJavaSession(void);
@@ -186,6 +187,7 @@ enum initstage
 	IS_FORMLESS_VOID,
 	IS_GUCS_REGISTERED,
 	IS_CAND_JVMLOCATION,
+	IS_CAND_POLICYURLS,
 	IS_PLJAVA_ENABLED,
 	IS_CAND_JVMOPENED,
 	IS_CREATEVM_SYM_FOUND,
@@ -194,6 +196,7 @@ enum initstage
 	IS_JAVAVM_STARTED,
 	IS_SIGHANDLERS,
 	IS_PLJAVA_FOUND,
+	IS_PLJAVA_INSTALLING,
 	IS_COMPLETE
 };
 
@@ -203,7 +206,10 @@ static bool jvmStartedAtLeastOnce = false;
 static bool alteredSettingsWereNeeded = false;
 static bool loadAsExtensionFailed = false;
 static bool seenVisualVMName;
+static bool seenModuleMain;
 static char const visualVMprefix[] = "-Dvisualvm.display.name=";
+static char const moduleMainPrefix[] = "-Djdk.module.main=";
+static char const policyUrlsGUC[] = "pljava.policy_urls";
 
 /*
  * In a background worker, _PG_init may be called very early, before much of
@@ -219,6 +225,28 @@ static char const visualVMprefix[] = "-Dvisualvm.display.name=";
  */
 static bool deferInit = false;
 
+/*
+ * Whether Backend_warnJEP411() should emit a warning when called.
+ * Initially true, because it may be called very early from the deferInit check,
+ * if pg_upgrade is happening, and should always warn in that case. Thereafter
+ * false, unless set true in the initsequencer because InstallHelper_groundwork
+ * will be called (PL/Java being installed or upgraded), or in the validator
+ * handler because a PL/Java function has been declared or redeclared.
+ */
+static bool warnJEP411 = true;
+
+/*
+ * Don't bother with the warning unless the JVM in use is later than Java 11.
+ * 11 is the LTS release prior to the one where JEP 411 gets interesting (17).
+ * If a site is sticking to LTS releases, there will be plenty of time to warn
+ * on 17. If a site moves with non-LTS releases, start warning as soon as
+ * anything > 11 is used.
+ *
+ * Initially true, so there will be a warning unconditionally in a case
+ * (pg_upgrade) where a JVM hasn't been launched to learn its version).
+ */
+static bool javaGT11 = true;
+
 static void initsequencer(enum initstage is, bool tolerant);
 
 #if PG_VERSION_NUM >= 90100
@@ -226,10 +254,14 @@ static void initsequencer(enum initstage is, bool tolerant);
 		char **newval, void **extra, GucSource source);
 	static bool check_vmoptions(
 		char **newval, void **extra, GucSource source);
-	static bool check_classpath(
+	static bool check_modulepath(
+		char **newval, void **extra, GucSource source);
+	static bool check_policy_urls(
 		char **newval, void **extra, GucSource source);
 	static bool check_enabled(
 		bool *newval, void **extra, GucSource source);
+	static bool check_java_thread_pg_entry(
+		int *newval, void **extra, GucSource source);
 
 	/* Check hooks will always allow "setting" a value that is the same as
 	 * current; otherwise, it would be frustrating to have just found settings
@@ -275,17 +307,36 @@ static void initsequencer(enum initstage is, bool tolerant);
 		return false;
 	}
 
-	static bool check_classpath(
+	static bool check_modulepath(
 		char **newval, void **extra, GucSource source)
 	{
 		if ( initstage < IS_JAVAVM_OPTLIST )
 			return true;
-		if ( classpath == *newval )
+		if ( modulepath == *newval )
 			return true;
-		if ( classpath && *newval && 0 == strcmp(classpath, *newval) )
+		if ( modulepath && *newval && 0 == strcmp(modulepath, *newval) )
 			return true;
 		GUC_check_errmsg(
-			"too late to change \"pljava.classpath\" setting");
+			"too late to change \"pljava.module_path\" setting");
+		GUC_check_errdetail(
+			"Changing the setting has no effect after "
+			"PL/Java has started the Java virtual machine.");
+		GUC_check_errhint(
+			"To try a different value, exit this session and start a new one.");
+		return false;
+	}
+
+	static bool check_policy_urls(
+		char **newval, void **extra, GucSource source)
+	{
+		if ( initstage < IS_JAVAVM_OPTLIST )
+			return true;
+		if ( policy_urls == *newval )
+			return true;
+		if ( policy_urls && *newval && 0 == strcmp(policy_urls, *newval) )
+			return true;
+		GUC_check_errmsg(
+			"too late to change \"pljava.policy_urls\" setting");
 		GUC_check_errdetail(
 			"Changing the setting has no effect after "
 			"PL/Java has started the Java virtual machine.");
@@ -309,6 +360,22 @@ static void initsequencer(enum initstage is, bool tolerant);
 			"For another chance, exit this session and start a new one.");
 		return false;
 	}
+
+	static bool check_java_thread_pg_entry(
+		int *newval, void **extra, GucSource source)
+	{
+		if ( initstage < IS_PLJAVA_FOUND )
+			return true;
+		if ( java_thread_pg_entry == *newval )
+			return true;
+		GUC_check_errmsg(
+			"too late to change \"pljava.java_thread_pg_entry\" setting");
+		GUC_check_errdetail(
+			"Start-up has progressed past the point where it is checked.");
+		GUC_check_errhint(
+			"For another chance, exit this session and start a new one.");
+		return false;
+	}
 #endif
 
 #if PG_VERSION_NUM < 90100
@@ -340,26 +407,17 @@ static void initsequencer(enum initstage is, bool tolerant);
 #define ASSIGNSTRINGHOOK(name) ASSIGNHOOK(name, const char *)
 #endif
 
-#if PG_VERSION_NUM >= 80400
 #define ASSIGNENUMHOOK(name) ASSIGNHOOK(name,int)
 #define ENUMBOOTVAL(entry) ((entry).val)
 #define ENUMHOOKRET true
-#else
-#define ASSIGNENUMHOOK(name) ASSIGNSTRINGHOOK(name)
-#define ENUMBOOTVAL(entry) ((char *)((entry).name))
-#define ENUMHOOKRET newval
-struct config_enum_entry
-{
-	const char *name;
-	int 		val;
-	bool		hidden;
-};
-#endif
 
 static const struct config_enum_entry java_thread_pg_entry_options[] = {
-	{"allow", 0, false},
-	{"error", 1, false},
-	{"block", 3, false},
+	{"allow", 0, false}, /* numeric value is bit-coded: */
+	{"error", 1, false}, /* 1: C code should refuse JNI calls on wrong thread */
+	                     /* 2: C code shouldn't call MonitorEnter/MonitorExit */
+	{"block", 3, false}, /* (3: check thread AND skip MonitorEnter/Exit)      */
+	                     /* 4: *Java* code should refuse wrong-thread calls   */
+	{"throw", 6, false}, /* (6: check in Java AND skip C MonitorEnter/Exit)   */
 	{NULL, 0, false}
 };
 
@@ -389,10 +447,10 @@ ASSIGNSTRINGHOOK(vmoptions)
 	ASSIGNRETURN(newval);
 }
 
-ASSIGNSTRINGHOOK(classpath)
+ASSIGNSTRINGHOOK(modulepath)
 {
 	ASSIGNRETURNIFCHECK(newval);
-	classpath = (char *)newval;
+	modulepath = (char *)newval;
 	if ( IS_FORMLESS_VOID < initstage && initstage < IS_JAVAVM_OPTLIST )
 	{
 		ASSIGNRETURNIFNXACT(newval);
@@ -402,6 +460,19 @@ ASSIGNSTRINGHOOK(classpath)
 	ASSIGNRETURN(newval);
 }
 
+ASSIGNSTRINGHOOK(policy_urls)
+{
+	ASSIGNRETURNIFCHECK(newval);
+	policy_urls = (char *)newval;
+	if ( IS_FORMLESS_VOID < initstage && initstage < IS_JAVAVM_OPTLIST )
+	{
+		alteredSettingsWereNeeded = true;
+		ASSIGNRETURNIFNXACT(newval);
+		initsequencer( initstage, true);
+	}
+	ASSIGNRETURN(newval);
+}
+
 ASSIGNHOOK(enabled, bool)
 {
 	ASSIGNRETURNIFCHECK(true);
@@ -417,21 +488,7 @@ ASSIGNHOOK(enabled, bool)
 
 ASSIGNENUMHOOK(java_thread_pg_entry)
 {
-#if PG_VERSION_NUM >= 80400
 	int val = newval;
-#else
-	int val = -1;
-	struct config_enum_entry const *e;
-	for ( e = java_thread_pg_entry_options; NULL != e->name; ++ e )
-	{
-		if ( 0 == strcmp(e->name, newval) )
-		{
-			val = e->val;
-		}
-	}
-	if ( -1 == val )
-		ASSIGNRETURN(NULL);
-#endif
 	ASSIGNRETURNIFCHECK(ENUMHOOKRET);
 	pljava_JNI_setThreadPolicy( !!(val&1) /*error*/, !(val&2) /*monitorops*/);
 	ASSIGNRETURN(ENUMHOOKRET);
@@ -458,6 +515,17 @@ ASSIGNENUMHOOK(java_thread_pg_entry)
  *    to succeed.
  * 3. From a GUC assign hook, if the user has updated a setting that might allow
  *    initialization to succeed. It resumes from where it left off.
+ * 4. From the validator handler, if initialization isn't complete yet. That
+ *    will definitely happen during pg_upgrade, which is a case where deferInit
+ *    will have been set. The validator will then clear deferInit and try to get
+ *    further in the init sequence. Importantly, pg_upgrade also sets
+ *    check_function_bodies false, which limits the validator's work to a syntax
+ *    check of the AS string. The validator therefore will not need to obtain a
+ *    schemaLoader or do anything else that requires the sqlj schema to be fully
+ *    populated (as, during pg_upgrade, it may not yet be). However, the
+ *    validator handler must avoid any action that sets pljavaLoadPath, as a
+ *    non-NULL value there would be treated below as case 1a, and trigger an
+ *    attempt to set up the sqlj schema.
  *
  * In all cases, the sequence must progress as far as starting the VM and
  * initializing the PL/Java classes. In all cases except 1a, that's enough,
@@ -484,6 +552,8 @@ static void initsequencer(enum initstage is, bool tolerant)
 		initstage = IS_GUCS_REGISTERED;
 		if ( deferInit )
 			return;
+		warnJEP411 = false;
+		/*FALLTHROUGH*/
 
 	case IS_GUCS_REGISTERED:
 		if ( NULL == libjvmlocation )
@@ -496,8 +566,22 @@ static void initsequencer(enum initstage is, bool tolerant)
 			goto check_tolerant;
 		}
 		initstage = IS_CAND_JVMLOCATION;
+		/*FALLTHROUGH*/
 
 	case IS_CAND_JVMLOCATION:
+		if ( NULL == policy_urls )
+		{
+			ereport(WARNING, (
+				errmsg("Java virtual machine not yet loaded"),
+				errdetail("Java policy URL(s) not configured"),
+				errhint("SET pljava.policy_urls TO the security policy "
+						"files PL/Java is to use.")));
+			goto check_tolerant;
+		}
+		initstage = IS_CAND_POLICYURLS;
+		/*FALLTHROUGH*/
+
+	case IS_CAND_POLICYURLS:
 		if ( ! pljavaEnabled )
 		{
 			ereport(WARNING, (
@@ -510,6 +594,7 @@ static void initsequencer(enum initstage is, bool tolerant)
 			goto check_tolerant;
 		}
 		initstage = IS_PLJAVA_ENABLED;
+		/*FALLTHROUGH*/
 
 	case IS_PLJAVA_ENABLED:
 		libjvm_handle = pg_dlopen(libjvmlocation);
@@ -523,6 +608,7 @@ static void initsequencer(enum initstage is, bool tolerant)
 			goto check_tolerant;
 		}
 		initstage = IS_CAND_JVMOPENED;
+		/*FALLTHROUGH*/
 
 	case IS_CAND_JVMOPENED:
 		pljava_createvm =
@@ -546,6 +632,7 @@ static void initsequencer(enum initstage is, bool tolerant)
 			goto check_tolerant;
 		}
 		initstage = IS_CREATEVM_SYM_FOUND;
+		/*FALLTHROUGH*/
 
 	case IS_CREATEVM_SYM_FOUND:
 		s_javaLogLevel = INFO;
@@ -559,26 +646,38 @@ static void initsequencer(enum initstage is, bool tolerant)
 		pljavaDebug = 1;
 #endif
 		initstage = IS_MISC_ONCE_DONE;
+		/*FALLTHROUGH*/
 
 	case IS_MISC_ONCE_DONE:
 		JVMOptList_init(&optList); /* uses CurrentMemoryContext */
 		seenVisualVMName = false;
+		seenModuleMain = false;
 		addUserJVMOptions(&optList);
 		if ( ! seenVisualVMName )
 			JVMOptList_addVisualVMName(&optList);
+		if ( ! seenModuleMain )
+			JVMOptList_addModuleMain(&optList);
 		JVMOptList_add(&optList, "vfprintf", (void*)my_vfprintf, true);
 #ifndef GCJ
 		JVMOptList_add(&optList, "-Xrs", 0, true);
 #endif
-		effectiveClassPath = getClassPath("-Djava.class.path=");
-		if(effectiveClassPath != 0)
+		effectiveModulePath = getModulePath("--module-path=");
+		if(effectiveModulePath != 0)
 		{
-			JVMOptList_add(&optList, effectiveClassPath, 0, true);
+			JVMOptList_add(&optList, effectiveModulePath, 0, true);
 		}
 		initstage = IS_JAVAVM_OPTLIST;
+		/*FALLTHROUGH*/
 
 	case IS_JAVAVM_OPTLIST:
+		/* Register an on_proc_exit handler that destroys the VM if it has
+		 * been started. It will also log a last-ditch message if the VM happens
+		 * to rudely call exit() rather than returning a non-OK result.
+		 */
+		on_proc_exit(_destroyJavaVM, 0);
+		s_startingVM = true;
 		JNIresult = initializeJavaVM(&optList); /* frees the optList */
+		s_startingVM = false;
 		if( JNI_OK != JNIresult )
 		{
 			initstage = IS_MISC_ONCE_DONE; /* optList has been freed */
@@ -599,6 +698,7 @@ static void initsequencer(enum initstage is, bool tolerant)
 		jvmStartedAtLeastOnce = true;
 		elog(DEBUG2, "successfully created Java virtual machine");
 		initstage = IS_JAVAVM_STARTED;
+		/*FALLTHROUGH*/
 
 	case IS_JAVAVM_STARTED:
 #ifdef USE_PLJAVA_SIGHANDLERS
@@ -606,10 +706,8 @@ static void initsequencer(enum initstage is, bool tolerant)
 		pqsignal(SIGTERM, pljavaDieHandler);
 		pqsignal(SIGQUIT, pljavaQuickDieHandler);
 #endif
-		/* Register an on_proc_exit handler that destroys the VM
-		 */
-		on_proc_exit(_destroyJavaVM, 0);
 		initstage = IS_SIGHANDLERS;
+		/*FALLTHROUGH*/
 
 	case IS_SIGHANDLERS:
 		Invocation_pushBootContext(&ctx);
@@ -647,17 +745,20 @@ static void initsequencer(enum initstage is, bool tolerant)
 		{
 			/* JVM initialization failed for some reason. Destroy
 			 * the VM if it exists. Perhaps the user will try
-			 * fixing the pljava.classpath and make a new attempt.
+			 * fixing the pljava.module_path and make a new attempt.
 			 */
 			ereport(WARNING, (
 				errmsg("failed to load initial PL/Java classes"),
-				errhint("The most common reason is that \"pljava.classpath\" "
-					"needs to be set, naming the proper \"pljava.jar\" file.")
+				errhint("The most common reason is that \"pljava.module_path\" "
+					"needs to be set, naming the proper \"pljava.jar\" "
+					"and \"pljava-api.jar\" files, separated by the correct "
+					"path separator for this platform.")
 					));
 			pljava_DualState_unregister();
 			_destroyJavaVM(0, 0);
 			goto check_tolerant;
 		}
+		/*FALLTHROUGH*/
 
 	case IS_PLJAVA_FOUND:
 		greeting = InstallHelper_hello();
@@ -665,9 +766,17 @@ static void initsequencer(enum initstage is, bool tolerant)
 				errmsg("PL/Java loaded"),
 				errdetail("versions:\n%s", greeting)));
 		pfree(greeting);
+		initstage = IS_PLJAVA_INSTALLING;
+		/*FALLTHROUGH*/
+
+	case IS_PLJAVA_INSTALLING:
 		if ( NULL != pljavaLoadPath )
+		{
+			warnJEP411 = javaGT11;
 			InstallHelper_groundwork(); /* sqlj schema, language handlers, ...*/
+		}
 		initstage = IS_COMPLETE;
+		/*FALLTHROUGH*/
 
 	case IS_COMPLETE:
 		pljavaLoadingAsExtension = false;
@@ -808,11 +917,7 @@ static void reLogWithChangedLevel(int level)
 	FreeErrorData(edata);
 #else
 	if (!errstart(level, edata->filename, edata->lineno,
-				  edata->funcname
-#if PG_VERSION_NUM >= 80400
-				  , NULL
-#endif
-				 ))
+				  edata->funcname, NULL))
 	{
 		FreeErrorData(edata);
 		return;
@@ -823,10 +928,8 @@ static void reLogWithChangedLevel(int level)
 		errmsg("%s", edata->message);
 	if (edata->detail)
 		errdetail("%s", edata->detail);
-#if PG_VERSION_NUM >= 80400
 	if (edata->detail_log)
 		errdetail_log("%s", edata->detail_log);
-#endif
 	if (edata->hint)
 		errhint("%s", edata->hint);
 	if (edata->context)
@@ -853,8 +956,27 @@ static void reLogWithChangedLevel(int level)
 
 void _PG_init()
 {
-	if ( IS_PLJAVA_FOUND == initstage )
+	char *sep;
+
+	if ( IS_PLJAVA_INSTALLING == initstage )
 		return; /* creating handler functions will cause recursive call */
+
+	InstallHelper_earlyHello();
+
+	/*
+	 * Find the platform's path separator. Java knows it, but that's no help in
+	 * preparing the launch options before it is launched. PostgreSQL knows what
+	 * it is, but won't directly say; give it some choices and it'll pick one.
+	 * Alternatively, let Maven or Ant determine and add a -D at build time from
+	 * the path.separator property. Maybe that's cleaner? This only works for
+	 * PG_VERSION_NUM >= 90100.
+	 */
+	sep = first_path_var_separator(":;");
+	if ( NULL == sep )
+		elog(ERROR,
+			"PL/Java cannot determine the path separator this platform uses");
+	s_path_var_sep = *sep;
+
 	if ( InstallHelper_shouldDeferInit() )
 		deferInit = true;
 	else
@@ -864,7 +986,7 @@ void _PG_init()
 
 static void initPLJavaClasses(void)
 {
-	jfieldID tlField;
+	jfieldID fID;
 	JNINativeMethod backendMethods[] =
 	{
 		{
@@ -902,6 +1024,31 @@ static void initPLJavaClasses(void)
 		"()Z",
 		Java_org_postgresql_pljava_internal_Backend__1isCreatingExtension
 		},
+		{
+		"_myLibraryPath",
+		"()Ljava/lang/String;",
+		Java_org_postgresql_pljava_internal_Backend__1myLibraryPath
+		},
+		{
+		"_pokeJEP411",
+		"(Ljava/lang/Class;Ljava/lang/Object;)V",
+		Java_org_postgresql_pljava_internal_Backend__1pokeJEP411
+		},
+		{ 0, 0, 0 }
+	};
+
+	JNINativeMethod earlyMethods[] =
+	{
+		{
+		"_forbidOtherThreads",
+		"()Z",
+		Java_org_postgresql_pljava_internal_Backend_00024EarlyNatives__1forbidOtherThreads
+		},
+		{
+		"_defineClass",
+		"(Ljava/lang/String;Ljava/lang/ClassLoader;[B)Ljava/lang/Class;",
+		Java_org_postgresql_pljava_internal_Backend_00024EarlyNatives__1defineClass
+		},
 		{ 0, 0, 0 }
 	};
 	jclass cls;
@@ -912,15 +1059,24 @@ static void initPLJavaClasses(void)
 
 	Exception_initialize();
 
-	elog(DEBUG2, "checking for a PL/Java Backend class on the given classpath");
+	elog(DEBUG2,
+		"checking for a PL/Java Backend class on the given module path");
+
+	cls = PgObject_getJavaClass(
+		"org/postgresql/pljava/internal/Backend$EarlyNatives");
+	PgObject_registerNatives2(cls, earlyMethods);
 
 	cls = PgObject_getJavaClass("org/postgresql/pljava/internal/Backend");
 	elog(DEBUG2, "successfully loaded Backend class");
 	s_Backend_class = JNI_newGlobalRef(cls);
 	PgObject_registerNatives2(s_Backend_class, backendMethods);
 
-	tlField = PgObject_getStaticJavaField(s_Backend_class, "THREADLOCK", "Ljava/lang/Object;");
-	JNI_setThreadLock(JNI_getStaticObjectField(s_Backend_class, tlField));
+	fID = PgObject_getStaticJavaField(s_Backend_class, "JAVA_MAJOR", "I");
+	javaGT11 = 11 < JNI_getStaticIntField(s_Backend_class, fID);
+
+	fID = PgObject_getStaticJavaField(s_Backend_class,
+		"THREADLOCK", "Ljava/lang/Object;");
+	JNI_setThreadLock(JNI_getStaticObjectField(s_Backend_class, fID));
 
 	Invocation_initialize();
 	Exception_initialize2();
@@ -937,33 +1093,6 @@ static void initPLJavaClasses(void)
 	SQLOutputToTuple_initialize();
 
 	InstallHelper_initialize();
-
-	s_setTrusted = PgObject_getStaticJavaMethod(s_Backend_class, "setTrusted", "(Z)V");
-}
-
-/**
- *  Initialize security
- */
-void Backend_setJavaSecurity(bool trusted)
-{
-	if(trusted != s_currentTrust)
-	{
-		/* GCJ has major issues here. Real work on SecurityManager and
-		 * related classes has just started in version 4.0.0.
-		 */
-#ifndef GCJ
-		JNI_callStaticVoidMethod(s_Backend_class, s_setTrusted, (jboolean)trusted);
-		if(JNI_exceptionCheck())
-		{
-			JNI_exceptionDescribe();
-			JNI_exceptionClear();
-			ereport(ERROR, (
-				errcode(ERRCODE_INTERNAL_ERROR),
-				errmsg("Unable to initialize java security")));
-		}
-#endif
-		s_currentTrust = trusted;
-	}
 }
 
 int Backend_setJavaLogLevel(int logLevel)
@@ -1086,30 +1215,32 @@ static void appendPathParts(const char* path, StringInfoData* bld, HashMap uniqu
 	for (;;)
 	{
 		char* pathPart;
+		char* sep;
 		size_t len;
+
 		if(*path == 0)
 			break;
 
-		len = strcspn(path, ";:");
+		sep = first_path_var_separator(path);
 
-		if(len == 1 && *(path+1) == ':' && isalnum(*path))
-			/*
-			 * Windows drive designator, leave it "as is".
-			 */
-			len = strcspn(path+2, ";:") + 2;
-		else
-		if(len == 0)
-			{
+		if(sep == path)
+		{
 			/* Ignore zero length components.
 			 */
 			++path;
 			continue;
-			}
+		}
+
+		if ( NULL == sep )
+			len = strlen(path);
+		else
+			len = sep - path;
 
 		initStringInfo(&buf);
 		if(*path == '$')
 		{
-			if(len == 7 || (strcspn(path, "/\\") == 7 && strncmp(path, "$libdir", 7) == 0))
+			if( (len == 7 || first_dir_separator(path) == path + 7)
+				&& strncmp(path, "$libdir", 7) == 0)
 			{
 				char pathbuf[MAXPGPATH];
 				get_pkglib_path(my_exec_path, pathbuf);
@@ -1120,7 +1251,8 @@ static void appendPathParts(const char* path, StringInfoData* bld, HashMap uniqu
 			else
 				ereport(ERROR, (
 					errcode(ERRCODE_INVALID_NAME),
-					errmsg("invalid macro name '%*s' in PL/Java classpath", (int)len, path)));
+					errmsg("invalid macro name '%*s' in PL/Java module path",
+						(int)len, path)));
 		}
 
 		if(len > 0)
@@ -1135,32 +1267,29 @@ static void appendPathParts(const char* path, StringInfoData* bld, HashMap uniqu
 			if(HashMap_size(unique) == 0)
 				appendStringInfo(bld, "%s", prefix);
 			else
-#if defined(WIN32)
-				appendStringInfoChar(bld, ';');
-#else
-				appendStringInfoChar(bld, ':');
-#endif
+				appendStringInfoChar(bld, s_path_var_sep);
 			appendStringInfo(bld, "%s", pathPart);
 			HashMap_putByString(unique, pathPart, (void*)1);
 		}
 		pfree(pathPart);
 		if(*path == 0)
 			break;
-		++path; /* Skip ':' */
+		++path; /* Skip path var separator */
 	}
 }
 
 /*
- * Get the CLASSPATH. Result is always freshly palloc'd.
+ * Get the module path. Result is always freshly palloc'd.
+ * No longer relies on an environment variable. What CLASSPATH variable might
+ * happen to be randomly set in the environment of a PostgreSQL backend?
  */
-static char* getClassPath(const char* prefix)
+static char* getModulePath(const char* prefix)
 {
 	char* path;
 	HashMap unique = HashMap_create(13, CurrentMemoryContext);
 	StringInfoData buf;
 	initStringInfo(&buf);
-	appendPathParts(classpath, &buf, unique, prefix);
-	appendPathParts(getenv("CLASSPATH"), &buf, unique, prefix);
+	appendPathParts(modulepath, &buf, unique, prefix);
 	PgObject_free((PgObject)unique);
 	path = buf.data;
 	if(strlen(path) == 0)
@@ -1232,7 +1361,25 @@ static void terminationTimeoutHandler(
  */
 static void _destroyJavaVM(int status, Datum dummy)
 {
-	if(s_javaVM != 0)
+	if(s_javaVM == 0)
+	{
+		if ( s_startingVM )
+		{
+			ereport(FATAL, (
+				errcode(ERRCODE_INTERNAL_ERROR),
+				errmsg("the Java VM exited while loading PL/Java"),
+				errdetail(
+					"The Java VM's exit forces this session to end."),
+				errhint(
+					"This has been known to happen when the entry in "
+					"pljava.module_path for the pljava-api jar has been "
+					"misspelled or the jar cannot be opened. If "
+					"logging_collector is active, there may be useful "
+					"information in the log.")
+					));
+		}
+	}
+	else
 	{
 		Invocation ctx;
 #ifdef USE_PLJAVA_SIGHANDLERS
@@ -1243,12 +1390,13 @@ static void _destroyJavaVM(int status, Datum dummy)
 		pqsigfunc saveSigAlrm;
 #endif
 
-		Invocation_pushInvocation(&ctx, false);
+		Invocation_pushBootContext(&ctx);
 		if(sigsetjmp(recoverBuf, 1) != 0)
 		{
 			elog(DEBUG2,
 				"needed to forcibly shut down the Java virtual machine");
 			s_javaVM = 0;
+			currentInvocation = 0;
 			return;
 		}
 
@@ -1270,7 +1418,7 @@ static void _destroyJavaVM(int status, Datum dummy)
 #endif
 
 #else
-		Invocation_pushInvocation(&ctx, false);
+		Invocation_pushBootContext(&ctx);
 		elog(DEBUG2, "shutting down the Java virtual machine");
 		JNI_destroyVM(s_javaVM);
 #endif
@@ -1324,6 +1472,9 @@ static void JVMOptList_add(JVMOptList* jol, const char* optString, void* extraIn
 	if ( 0 == strncmp(optString, visualVMprefix, sizeof visualVMprefix - 1) )
 		seenVisualVMName = true;
 
+	if ( 0 == strncmp(optString, moduleMainPrefix, sizeof moduleMainPrefix-1) )
+		seenModuleMain = true;
+
 	elog(DEBUG2, "Added JVM option string \"%s\"", optString);
 }
 
@@ -1341,6 +1492,15 @@ static void JVMOptList_addVisualVMName(JVMOptList* jol)
 	JVMOptList_add(jol, buf.data, 0, false);
 }
 
+static void JVMOptList_addModuleMain(JVMOptList* jol)
+{
+	StringInfoData buf;
+	initStringInfo(&buf);
+	appendStringInfo(&buf, "%s%s",
+		moduleMainPrefix, "org.postgresql.pljava");
+	JVMOptList_add(jol, buf.data, 0, false);
+}
+
 /* Split JVM options. The string is split on whitespace unless the
  * whitespace is found within a string or is escaped by backslash. A
  * backslash escaped quote is not considered a string delimiter.
@@ -1460,7 +1620,7 @@ static jint initializeJavaVM(JVMOptList *optList)
 
 	vm_args.nOptions = optList->size;
 	vm_args.options  = optList->options;
-	vm_args.version  = JNI_VERSION_1_4;
+	vm_args.version  = JNI_VERSION_9;
 	vm_args.ignoreUnrecognized = JNI_FALSE;
 
 	elog(DEBUG2, "creating Java virtual machine");
@@ -1478,21 +1638,9 @@ static jint initializeJavaVM(JVMOptList *optList)
 	return jstat;
 }
 
-#if PG_VERSION_NUM >= 80400
 #define GUCBOOTVAL(v) (v),
 #define GUCBOOTASSIGN(a, v)
 #define GUCFLAGS(f) (f),
-#else
-#define GUCBOOTVAL(v)
-#define GUCBOOTASSIGN(a, v) \
-	StaticAssertStmt(NULL != (valueAddr), "NULL valueAddr for GUC"); \
-	*(a) = (v);
-#define GUCFLAGS(f)
-#define DefineCustomEnumVariable(name, short_desc, long_desc, valueAddr, \
-		options, context, assign_hook, show_hook) \
-	DefineCustomStringVariable((name), (short_desc), (long_desc), (valueAddr), \
-		(context), (assign_hook), (show_hook))
-#endif
 
 #if PG_VERSION_NUM >= 90100
 #define GUCCHECK(h) (h),
@@ -1555,7 +1703,7 @@ static void registerGUCOptions(void)
 		&libjvmlocation,
 		PLJAVA_LIBJVMDEFAULT,
 		PGC_SUSET,
-		0,    /* flags */
+		GUC_SUPERUSER_ONLY,    /* flags */
 		check_libjvm_location,
 		assign_libjvm_location,
 		NULL); /* show hook */
@@ -1567,21 +1715,39 @@ static void registerGUCOptions(void)
 		&vmoptions,
 		NULL, /* boot value */
 		PGC_SUSET,
-		0,    /* flags */
+		GUC_SUPERUSER_ONLY,    /* flags */
 		check_vmoptions,
 		assign_vmoptions,
 		NULL); /* show hook */
 
 	STRING_GUC(
-		"pljava.classpath",
-		"Classpath used by the JVM",
+		"pljava.module_path",
+		"Module path to be used by the JVM",
 		NULL, /* extended description */
-		&classpath,
-		InstallHelper_defaultClassPath(pathbuf), /* boot value */
+		&modulepath,
+		InstallHelper_defaultModulePath(pathbuf,s_path_var_sep),/* boot value */
 		PGC_SUSET,
-		0,    /* flags */
-		check_classpath,
-		assign_classpath,
+		GUC_SUPERUSER_ONLY,    /* flags */
+		check_modulepath,
+		assign_modulepath,
+		NULL); /* show hook */
+
+	STRING_GUC(
+		policyUrlsGUC,
+		"URLs to Java security policy file(s) for PL/Java's use",
+		"Quote each URL and separate with commas. Any URL may begin (inside "
+		"the quotes) with n= where n is the index of the Java "
+		"policy.url.n property to set. If not specified, the first will "
+		"become policy.url.2 (following the JRE-installed policy) with "
+		"subsequent entries following in sequence. The last entry may be a "
+		"bare = (still quoted) to prevent use of any higher-numbered policy "
+		"URLs from the java.security file.",
+		&policy_urls,
+		"\"file:${org.postgresql.sysconfdir}/pljava.policy\",\"=\"",
+		PGC_SUSET,
+		PLJAVA_IMPLEMENTOR_FLAGS | GUC_SUPERUSER_ONLY,
+		check_policy_urls, /* check hook */
+		assign_policy_urls,
 		NULL); /* show hook */
 
 	BOOL_GUC(
@@ -1651,13 +1817,15 @@ static void registerGUCOptions(void)
 		"entered Java. If 'error', any thread other than the main one will "
 		"incur an exception if it tries to enter PG. If 'block', the main "
 		"thread will never release its lock, so any other thread that tries "
-		"to enter PG will indefinitely block.",
+		"to enter PG will indefinitely block. If 'throw', like 'error', other "
+		"threads will incur an exception, but earlier: it will be thrown "
+		"in Java, before the JNI boundary into C is even crossed.",
 		&java_thread_pg_entry,
 		ENUMBOOTVAL(java_thread_pg_entry_options[0]), /* allow */
 		java_thread_pg_entry_options,
 		PGC_USERSET,
 		0, /* flags */
-		NULL, /* check hook */
+		check_java_thread_pg_entry, /* check hook */
 		assign_java_thread_pg_entry,
 		NULL); /* display hook */
 
@@ -1675,7 +1843,7 @@ static void registerGUCOptions(void)
 #undef PLJAVA_ENABLE_DEFAULT
 #undef PLJAVA_IMPLEMENTOR_FLAGS
 
-static Datum internalCallHandler(bool trusted, PG_FUNCTION_ARGS);
+static inline Datum internalCallHandler(bool trusted, PG_FUNCTION_ARGS);
 
 extern PLJAVADLLEXPORT Datum javau_call_handler(PG_FUNCTION_ARGS);
 PG_FUNCTION_INFO_V1(javau_call_handler);
@@ -1699,10 +1867,13 @@ Datum java_call_handler(PG_FUNCTION_ARGS)
 	return internalCallHandler(true, fcinfo);
 }
 
-static Datum internalCallHandler(bool trusted, PG_FUNCTION_ARGS)
+static inline Datum
+internalCallHandler(bool trusted, PG_FUNCTION_ARGS)
 {
 	Invocation ctx;
 	Datum retval = 0;
+	Oid funcoid = fcinfo->flinfo->fn_oid;
+	bool forTrigger = CALLED_AS_TRIGGER(fcinfo);
 
 	/*
 	 * Just in case it could be helpful in offering diagnostics later, hang
@@ -1710,34 +1881,18 @@ static Datum internalCallHandler(bool trusted, PG_FUNCTION_ARGS)
 	 * It's cheap, and can be followed back to the right language and
 	 * handler function entries later if needed.
 	 */
-	*(trusted ? &pljavaTrustedOid : &pljavaUntrustedOid)
-		= fcinfo->flinfo->fn_oid;
+	*(trusted ? &pljavaTrustedOid : &pljavaUntrustedOid) = funcoid;
 	if ( IS_COMPLETE != initstage )
 	{
 		deferInit = false;
 		initsequencer( initstage, false);
-
-		/* Force initial setting
- 		 */
-		s_currentTrust = !trusted;
 	}
 
-	Invocation_pushInvocation(&ctx, trusted);
+	Invocation_pushInvocation(&ctx);
 	PG_TRY();
 	{
-		Function function = Function_getFunction(fcinfo);
-		if(CALLED_AS_TRIGGER(fcinfo))
-		{
-			/* Called as a trigger procedure
-			 */
-			retval = Function_invokeTrigger(function, fcinfo);
-		}
-		else
-		{
-			/* Called as a function
-			 */
-			retval = Function_invoke(function, fcinfo);
-		}
+		retval = Function_invoke(
+			funcoid, trusted, forTrigger, false, true, fcinfo);
 		Invocation_popInvocation(false);
 	}
 	PG_CATCH();
@@ -1749,12 +1904,148 @@ static Datum internalCallHandler(bool trusted, PG_FUNCTION_ARGS)
 	return retval;
 }
 
+static Datum internalValidator(bool trusted, PG_FUNCTION_ARGS);
+
+extern PLJAVADLLEXPORT Datum javau_validator(PG_FUNCTION_ARGS);
+PG_FUNCTION_INFO_V1(javau_validator);
+
+Datum javau_validator(PG_FUNCTION_ARGS)
+{
+	return internalValidator(false, fcinfo);
+}
+
+extern PLJAVADLLEXPORT Datum java_validator(PG_FUNCTION_ARGS);
+PG_FUNCTION_INFO_V1(java_validator);
+
+Datum java_validator(PG_FUNCTION_ARGS)
+{
+	return internalValidator(true, fcinfo);
+}
+
+static Datum internalValidator(bool trusted, PG_FUNCTION_ARGS)
+{
+	Oid funcoid = PG_GETARG_OID(0);
+	Invocation ctx;
+	Oid *oidSaveLocation = NULL;
+
+	if (!CheckFunctionValidatorAccess(fcinfo->flinfo->fn_oid, funcoid))
+		PG_RETURN_VOID();
+
+	/*
+	 * In the call handler, which could be called heavily, funcoid gets
+	 * unconditionally stored to one of these two locations, rather than
+	 * spending extra cycles deciding whether to store it or not. A validator
+	 * will not be called as heavily, and can afford to check here whether
+	 * an Oid needs to be stored or not. The situation to avoid is where
+	 * funcoid gets stored here, as an Oid from which PL/Java's library path can
+	 * be found, but the function then gets rejected by the validator, leaving
+	 * the stored Oid invalid and useless for that purpose. Therefore, choose
+	 * here whether and where to store it, but store it only within the PG_TRY
+	 * block, and replace with InvalidOid again in the PG_CATCH.
+	 */
+	if ( trusted )
+	{
+		if ( InvalidOid == pljavaTrustedOid )
+			oidSaveLocation = &pljavaTrustedOid;
+	}
+	else
+	{
+		if ( InvalidOid == pljavaUntrustedOid )
+			oidSaveLocation = &pljavaUntrustedOid;
+	}
+
+	if ( IS_PLJAVA_INSTALLING > initstage )
+	{
+		if ( check_function_bodies ) /* We're gonna need a JVM */
+		{
+			deferInit = false;
+			initsequencer( initstage, false);
+		}
+		else /* Can try to start one, but if no go, just assume function's ok */
+		{
+			initsequencer( initstage, true);
+			if ( IS_PLJAVA_INSTALLING > initstage )
+			{
+				if ( javaGT11 )
+					warnJEP411 = true;
+				PG_RETURN_VOID();
+			}
+		}
+	}
+
+	Invocation_pushInvocation(&ctx);
+	PG_TRY();
+	{
+		if ( NULL != oidSaveLocation )
+			*oidSaveLocation = funcoid;
+
+		Function_invoke(
+			funcoid, trusted, false, true, check_function_bodies, NULL);
+		Invocation_popInvocation(false);
+	}
+	PG_CATCH();
+	{
+		if ( NULL != oidSaveLocation )
+			*oidSaveLocation = InvalidOid;
+
+		Invocation_popInvocation(true);
+		PG_RE_THROW();
+	}
+	PG_END_TRY();
+
+	if ( javaGT11 )
+		warnJEP411 = true;
+	PG_RETURN_VOID();
+}
+
+/*
+ * Called at the ends of committing transactions to emit a warning about future
+ * JEP 411 impacts, at most once per session, if any PL/Java functions were
+ * declared or redeclared in the transaction, or if PL/Java was installed or
+ * upgraded. Also called from InstallHelper, if pg_upgrade is happening.
+ * Yes, this is a bit tangled. The tracking of function declaration happens
+ * above in the validator handler, and PL/Java installation/upgrade is detected
+ * in the initsequencer.
+ */
+void Backend_warnJEP411(bool isCommit)
+{
+	static bool warningEmitted = false; /* once only per session */
+
+	if ( warningEmitted  ||  ! warnJEP411 )
+		return;
+
+	if ( ! isCommit )
+	{
+		warnJEP411 = false;
+		return;
+	}
+
+	warningEmitted = true;
+
+	ereport(WARNING, (
+		errmsg(
+			"[JEP 411] migration advisory: there will be a Java version "
+			"(after Java 17) that will be unable to run PL/Java %s "
+			 "with policy enforcement", SO_VERSION_STRING),
+		errdetail(
+			"This PL/Java version enforces security policy using important "
+			"Java features that will be phased out in future Java versions. "
+			"Those changes will come in releases after Java 17."),
+		errhint(
+			"For migration planning, Java versions up to and including 17 "
+			"remain fully usable with this version of PL/Java, and Java 17 "
+			"is positioned as a long-term support release. For details on "
+			"how PL/Java will adapt, please bookmark "
+			"https://github.com/tada/pljava/wiki/JEP-411")
+	));
+}
+
 /****************************************
  * JNI methods
  ****************************************/
 JNIEXPORT jint JNICALL JNI_OnLoad(JavaVM* vm, void* reserved)
 {
-	return JNI_VERSION_1_4;
+	return JNI_VERSION_9;
 }
 
 /*
@@ -1773,7 +2064,11 @@ JNICALL Java_org_postgresql_pljava_internal_Backend__1getConfigOption(JNIEnv* en
 	{
 		PG_TRY();
 		{
-			const char *value = PG_GETCONFIGOPTION(key);
+			const char *value;
+			if ( 0 == strcmp(policyUrlsGUC, key) )
+				value = policy_urls;
+			else
+				value = PG_GETCONFIGOPTION(key);
 			pfree(key);
 			if(value != 0)
 				result = String_createJavaStringFromNTS(value);
@@ -1873,3 +2168,133 @@ Java_org_postgresql_pljava_internal_Backend__1isCreatingExtension(JNIEnv *env, j
 	pljavaCheckExtension( &inExtension);
 	return inExtension ? JNI_TRUE : JNI_FALSE;
 }
+
+/*
+ * Class:     org_postgresql_pljava_internal_Backend
+ * Method:    _myLibraryPath
+ * Signature: ()Ljava/lang/String;
+ */
+JNIEXPORT jstring JNICALL
+Java_org_postgresql_pljava_internal_Backend__1myLibraryPath(JNIEnv *env, jclass cls)
+{
+	jstring result = NULL;
+
+	BEGIN_NATIVE
+
+	if ( NULL == pljavaLoadPath )
+	{
+		Oid funcoid = pljavaTrustedOid;
+
+		if ( InvalidOid == funcoid )
+			funcoid = pljavaUntrustedOid;
+		if ( InvalidOid == funcoid )
+			return NULL;
+
+		/*
+		 * Result not needed, but pljavaLoadPath is set as a side effect.
+		 */
+		InstallHelper_isPLJavaFunction(funcoid, NULL, NULL);
+	}
+
+	if ( NULL != pljavaLoadPath )
+		result = String_createJavaStringFromNTS(pljavaLoadPath);
+
+	END_NATIVE
+
+	return result;
+}
+
+/*
+ * Class:     org_postgresql_pljava_internal_Backend
+ * Method:    _pokeJEP411
+ * Signature: (Ljava/lang/Class;Ljava/lang/Object;)V
+ *
+ * This method is hideously dependent on unexposed JDK internals. But then,
+ * the fact that it's needed at all is hideous already. Java, any language,
+ * is classic infrastructure. Other layers, like this, are built atop it, and
+ * others in turn use those layers. The idea that the language developers would
+ * arrogate to themselves the act of sending an inappropriately low-level
+ * message directly to ultimate users, insisting that the stack layers above
+ * cannot intercept it and notify the higher-level users in terms that fit
+ * the abstractions meaningful there, leaves an uneasy picture of how
+ * a development team can begin to lose sight of who is providing what to whom
+ * and why.
+ *
+ * At least as of the time of this writing, System has a CallersHolder class
+ * holding a map recording classes for which the warning has already been sent.
+ * Poking the 'caller' class into that map works to suppress the warning.
+ */
+JNIEXPORT void JNICALL
+Java_org_postgresql_pljava_internal_Backend__1pokeJEP411(JNIEnv *env, jclass cls, jclass caller, jobject token)
+{
+	jclass callersHolder;
+	jfieldID callers;
+	jobject map;
+	jclass mapClass;
+	jmethodID put;
+
+	BEGIN_NATIVE
+
+	callersHolder = JNI_findClass("java/lang/System$CallersHolder");
+	if ( NULL == callersHolder )
+		goto failed;
+
+	callers = JNI_getStaticFieldID(callersHolder, "callers", "Ljava/util/Map;");
+	if ( NULL == callers )
+		goto failed;
+
+	map = JNI_getStaticObjectField(callersHolder, callers);
+	if ( NULL == map )
+		goto failed;
+
+	mapClass = JNI_getObjectClass(map);
+	put = JNI_getMethodID(mapClass,
+		"put", "(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;");
+
+	JNI_callObjectMethodLocked(map, put, caller, token);
+	goto done;
+
+failed:
+	JNI_exceptionClear();
+
+done:
+	END_NATIVE
+}
+
+/*
+ * Class:     org_postgresql_pljava_internal_Backend_EarlyNatives
+ * Method:    _forbidOtherThreads
+ * Signature: ()Z
+ */
+JNIEXPORT jboolean JNICALL
+Java_org_postgresql_pljava_internal_Backend_00024EarlyNatives__1forbidOtherThreads(JNIEnv *env, jclass cls)
+{
+	return (java_thread_pg_entry & 4) ? JNI_TRUE : JNI_FALSE;
+}
+
+/*
+ * Class:     org_postgresql_pljava_internal_Backend_EarlyNatives
+ * Method:    _defineClass
+ * Signature: (Ljava/lang/String;Ljava/lang/ClassLoader;[B)Ljava/lang/Class;
+ */
+JNIEXPORT jclass JNICALL
+Java_org_postgresql_pljava_internal_Backend_00024EarlyNatives__1defineClass(JNIEnv *env, jclass cls, jstring name, jobject loader, jbyteArray image)
+{
+	const char *utfName;
+	jbyte *bytes;
+	jsize nbytes;
+	jclass newcls;
+	static bool oneShot = false;
+
+	if ( oneShot )
+		return NULL;
+	oneShot = true;
+
+	utfName = (*env)->GetStringUTFChars(env, name, NULL);
+	bytes = (*env)->GetByteArrayElements(env, image, NULL);
+	nbytes = (*env)->GetArrayLength(env, image);
+	newcls = (*env)->DefineClass(env, utfName, loader, bytes, nbytes);
+	(*env)->ReleaseByteArrayElements(env, image, bytes, JNI_ABORT);
+	(*env)->ReleaseStringUTFChars(env, name, utfName);
+	return newcls;
+}
