@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2015-2019 Tada AB and other contributors, as listed below.
+ * Copyright (c) 2015-2021 Tada AB and other contributors, as listed below.
  *
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the The BSD 3-Clause License
@@ -25,6 +25,7 @@
 #include <executor/spi.h>
 #include <miscadmin.h>
 #include <libpq/libpq-be.h>
+#include <postmaster/autovacuum.h>
 #include <tcop/pquery.h>
 #include <utils/builtins.h>
 #include <utils/lsyscache.h>
@@ -84,13 +85,11 @@
  * for 9.3.0 through 9.3.2.
  *
  * One thing it's needed for is to avoid dereferencing MyProcPort in a
- * background worker, where it's not set. Define BGW_HAS_NO_MYPROCPORT if that
- * has to be (and can be) checked.
+ * background worker, where it's not set.
  */
 #if PG_VERSION_NUM < 90300  ||  defined(_MSC_VER) && PG_VERSION_NUM < 90303
 #define IsBackgroundWorker false
 #else
-#define BGW_HAS_NO_MYPROCPORT
 #include <commands/dbcommands.h>
 #if defined(_MSC_VER)
 #include <postmaster/bgworker.h>
@@ -98,12 +97,6 @@
 #endif
 #endif
 
-#ifndef PLJAVA_SO_VERSION
-#error "PLJAVA_SO_VERSION needs to be defined to compile this file."
-#else
-#define SO_VERSION_STRING CppAsString2(PLJAVA_SO_VERSION)
-#endif
-
 /*
  * The name of the table the extension scripts will create to pass information
  * here. The table name is phrased as an error message because it will appear
@@ -115,11 +108,12 @@
 static jclass s_InstallHelper_class;
 static jmethodID s_InstallHelper_hello;
 static jmethodID s_InstallHelper_groundwork;
+static jfieldID  s_InstallHelper_MANAGE_CONTEXT_LOADER;
 
 static bool extensionExNihilo = false;
 
 static void checkLoadPath( bool *livecheck);
-static void getExtensionLoadPath();
+static void getExtensionLoadPath(void);
 static char *origUserName();
 
 char const *pljavaLoadPath = NULL;
@@ -137,11 +131,10 @@ bool pljavaViableXact()
 
 char *pljavaDbName()
 {
-#ifdef BGW_HAS_NO_MYPROCPORT
-	char *shortlived;
-	static char *longlived;
-	if ( IsBackgroundWorker )
+	if ( IsAutoVacuumWorkerProcess() || IsBackgroundWorker )
 	{
+		char *shortlived;
+		static char *longlived;
 		if ( NULL == longlived )
 		{
 			shortlived = get_database_name(MyDatabaseId);
@@ -153,14 +146,12 @@ char *pljavaDbName()
 		}
 		return longlived;
 	}
-#endif
 	return MyProcPort->database_name;
 }
 
 static char *origUserName()
 {
-#ifdef BGW_HAS_NO_MYPROCPORT
-	if ( IsBackgroundWorker )
+	if ( IsAutoVacuumWorkerProcess() || IsBackgroundWorker )
 	{
 #if PG_VERSION_NUM >= 90500
 		char *shortlived;
@@ -175,13 +166,12 @@ static char *origUserName()
 #else
 		ereport(ERROR, (
 			errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-			errmsg("PL/Java in a background worker not supported "
+			errmsg("PL/Java in a background or autovacuum worker not supported "
 				"in this PostgreSQL version"),
-			errhint("PostgreSQL 9.5 is the first version to support "
-				"PL/Java in a background worker.")));
+			errhint("PostgreSQL 9.5 is the first version in which "
+				"such usage is supported.")));
 #endif
 	}
-#endif
 	return MyProcPort->user_name;
 }
 
@@ -354,7 +344,7 @@ static void getExtensionLoadPath()
  *
  * If a string is returned, it has been palloc'd in the current context.
  */
-char *pljavaFnOidToLibPath(Oid fnOid)
+char *pljavaFnOidToLibPath(Oid fnOid, char **langName, bool *trusted)
 {
 	bool isnull;
 	HeapTuple procTup;
@@ -394,13 +384,15 @@ char *pljavaFnOidToLibPath(Oid fnOid)
 		elog(ERROR, "cache lookup failed for language %u", langId);
 	langStruct = (Form_pg_language) GETSTRUCT(langTup);
 	handlerOid = langStruct->lanplcallfoid;
-	ReleaseSysCache(langTup);
 	/*
 	 * PL/Java has certainly got a function call handler, so if this language
 	 * hasn't, PL/Java it's not.
 	 */
 	if ( InvalidOid == handlerOid )
+	{
+		ReleaseSysCache(langTup);
 		return NULL;
+	}
 
 	/*
 	 * Da capo al coda ... handlerOid is another function to be looked up.
@@ -413,7 +405,10 @@ char *pljavaFnOidToLibPath(Oid fnOid)
 	 * If the call handler's not a C function, this isn't PL/Java....
 	 */
 	if ( ClanguageId != procStruct->prolang )
+	{
+		ReleaseSysCache(langTup);
 		return NULL;
+	}
 
 	/*
 	 * Now that the handler is known to be a C function, it should have a
@@ -423,6 +418,11 @@ char *pljavaFnOidToLibPath(Oid fnOid)
 		SysCacheGetAttr(PROCOID, procTup, Anum_pg_proc_probin, &isnull);
 	if ( isnull )
 		elog(ERROR, "null probin for C function %u", handlerOid);
+	if ( NULL != langName )
+		*langName = pstrdup(NameStr(langStruct->lanname));
+	if ( NULL != trusted )
+		*trusted = langStruct->lanpltrusted;
+	ReleaseSysCache(langTup);
 	probinstring = /* TextDatumGetCString(probinattr); */
 		DatumGetCString(DirectFunctionCall1(textout, probinattr)); /*archaic*/
 	ReleaseSysCache(procTup);
@@ -440,16 +440,23 @@ char *pljavaFnOidToLibPath(Oid fnOid)
 
 bool InstallHelper_shouldDeferInit()
 {
-	return IsBackgroundWorker || IsBinaryUpgrade;
+	if ( IsBackgroundWorker || IsAutoVacuumWorkerProcess() )
+		return true;
+
+	if ( ! IsBinaryUpgrade )
+		return false;
+
+	Backend_warnJEP411(true);
+	return true;
 }
 
-bool InstallHelper_isPLJavaFunction(Oid fn)
+bool InstallHelper_isPLJavaFunction(Oid fn, char **langName, bool *trusted)
 {
 	char *itsPath;
 	char *pljPath;
 	bool result = false;
 
-	itsPath = pljavaFnOidToLibPath(fn);
+	itsPath = pljavaFnOidToLibPath(fn, langName, trusted);
 	if ( NULL == itsPath )
 		return false;
 
@@ -457,9 +464,9 @@ bool InstallHelper_isPLJavaFunction(Oid fn)
 	{
 		pljPath = NULL;
 		if ( InvalidOid != pljavaTrustedOid )
-			pljPath = pljavaFnOidToLibPath(pljavaTrustedOid);
+			pljPath = pljavaFnOidToLibPath(pljavaTrustedOid, NULL, NULL);
 		if ( NULL == pljPath && InvalidOid != pljavaUntrustedOid )
-			pljPath = pljavaFnOidToLibPath(pljavaUntrustedOid);
+			pljPath = pljavaFnOidToLibPath(pljavaUntrustedOid, NULL, NULL);
 		if ( NULL == pljPath )
 		{
 			elog(WARNING, "unable to determine PL/Java's load path");
@@ -475,30 +482,46 @@ finally:
 	return result;
 }
 
-char const *InstallHelper_defaultClassPath(char *pathbuf)
+char const *InstallHelper_defaultModulePath(char *pathbuf, char pathsep)
 {
 	char * const pbend = pathbuf + MAXPGPATH;
 	char *pbp = pathbuf;
 	size_t remaining;
-	size_t verlen = strlen(SO_VERSION_STRING);
+	int would_have_sprinted;
 
 	get_share_path(my_exec_path, pathbuf);
 	join_path_components(pathbuf, pathbuf, "pljava");
-	join_path_components(pathbuf, pathbuf, "pljava-");
+	join_path_components(pathbuf, pathbuf, "pljava"); /* puts \0 where - goes */
 
 	for ( ; pbp < pbend && '\0' != *pbp ; ++ pbp )
 		;
 	if ( pbend == pbp )
 		return NULL;
 
-	remaining = pbend - pbp;
-	if ( remaining < verlen + 5 )
+	/*
+	 * pbp now points to a \0 that should later be replaced with a hyphen.
+	 * The \0-terminated string starting at pathbuf can, for now, be reused
+	 * as an argument to snprintf.
+	 */
+
+	remaining = (pbend - pbp) - 1;
+
+	would_have_sprinted = snprintf(pbp + 1, remaining, "%s.jar%c%s-api-%s.jar",
+		SO_VERSION_STRING, pathsep, pathbuf, SO_VERSION_STRING);
+
+	if ( would_have_sprinted >= remaining )
 		return NULL;
 
-	snprintf(pbp, remaining, "%s.jar", SO_VERSION_STRING);
+	*pbp = '-'; /* overwrite the \0 so now it's a single string. */
 	return pathbuf;
 }
 
+void InstallHelper_earlyHello()
+{
+	elog(DEBUG2,
+		"pljava-so-" SO_VERSION_STRING " built for (" PG_VERSION_STR ")");
+}
+
 char *InstallHelper_hello()
 {
 	char pathbuf[MAXPGPATH];
@@ -523,6 +546,10 @@ char *InstallHelper_hello()
 	jstring greeting;
 	char *greetingC;
 	char const *clusternameC = pljavaClusterName();
+	jboolean manageContext = JNI_getStaticBooleanField(s_InstallHelper_class,
+		s_InstallHelper_MANAGE_CONTEXT_LOADER);
+
+	pljava_JNI_threadInitialize(JNI_TRUE == manageContext);
 
 	Invocation_pushBootContext(&ctx);
 	nativeVer = String_createJavaStringFromNTS(SO_VERSION_STRING);
@@ -585,7 +612,7 @@ void InstallHelper_groundwork()
 {
 	Invocation ctx;
 	bool snapshot_set = false;
-	Invocation_pushInvocation(&ctx, false);
+	Invocation_pushInvocation(&ctx);
 	ctx.function = Function_INIT_WRITER;
 #if PG_VERSION_NUM >= 80400
 	if ( ! ActiveSnapshotSet() )
@@ -645,6 +672,8 @@ void InstallHelper_initialize()
 {
 	s_InstallHelper_class = (jclass)JNI_newGlobalRef(PgObject_getJavaClass(
 		"org/postgresql/pljava/internal/InstallHelper"));
+	s_InstallHelper_MANAGE_CONTEXT_LOADER = PgObject_getStaticJavaField(
+		s_InstallHelper_class, "MANAGE_CONTEXT_LOADER", "Z");
 	s_InstallHelper_hello = PgObject_getStaticJavaMethod(s_InstallHelper_class,
 		"hello",
 		"(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;"
