@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004-2020 Tada AB and other contributors, as listed below.
+ * Copyright (c) 2004-2021 Tada AB and other contributors, as listed below.
  *
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the The BSD 3-Clause License
@@ -107,6 +107,7 @@ static jmethodID s_TypeBridge_Holder_payload;
 typedef struct
 {
 	Type          elemType;
+	Function      fn;
 	jobject       rowProducer;
 	jobject       rowCollector;
 	/*
@@ -126,18 +127,12 @@ typedef struct
 	 */
 	MemoryContext spiContext;
 	bool          hasConnected;
-	/*
-	 * Copy of Invocation's 'trusted' flag. In normal calls through the handler,
-	 * the value is known (implicit in which handler entry point was called),
-	 * but that isn't available to the _endOfSetCB, so must be remembered here.
-	 */
-	bool          trusted;
 } CallContextData;
 
 /*
  * Called during evaluation of a set-returning function, at various points after
  * calls into Java code could have instantiated an Invocation, or connected SPI.
- * Does not stash elemType, rowProducer, rowCollector, or trusted; those are all
+ * Does not stash elemType, rowProducer, or rowCollector; those are all
  * unconditionally set in the first-call initialization, and spiContext to zero.
  */
 static void stashCallContext(CallContextData *ctxData)
@@ -165,10 +160,23 @@ static void stashCallContext(CallContextData *ctxData)
  */
 static void _closeIteration(CallContextData* ctxData)
 {
+	jobject dummy;
 	currentInvocation->hasConnected = ctxData->hasConnected;
 	currentInvocation->invocation   = ctxData->invocation;
 
-	Type_closeSRF(ctxData->elemType, ctxData->rowProducer);
+	/*
+	 * Why pass 1 as the call_cntr? We won't always have the actual call_cntr
+	 * value at _closeIteration time (the _endOfSetCB isn't passed it), and the
+	 * Java interfaces being used don't need it (close() isn't passed a row
+	 * number), but vpcInvoke needs a way to know when to skip its call to
+	 * installContextLoader (which shouldn't happen on its very first call,
+	 * the one with call_cntr of zero, which always happens as part of the first
+	 * invocation of the SRF so the work has already been done). So passing any
+	 * fixed value here that isn't zero is enough to distinguish the cases.
+	 */
+	pljava_Function_vpcInvoke(
+		ctxData->fn, ctxData->rowProducer, NULL, 1, JNI_TRUE, &dummy);
+
 	JNI_deleteGlobalRef(ctxData->rowProducer);
 	if(ctxData->rowCollector != 0)
 		JNI_deleteGlobalRef(ctxData->rowCollector);
@@ -198,16 +206,31 @@ static void _closeIteration(CallContextData* ctxData)
  */
 static void _endOfSetCB(Datum arg)
 {
-	Invocation topCall;
-	bool saveInExprCtxCB;
+	Invocation ctx;
 	CallContextData* ctxData = (CallContextData*)DatumGetPointer(arg);
-	if(currentInvocation == 0)
-		Invocation_pushInvocation(&topCall, ctxData->trusted);
 
-	saveInExprCtxCB = currentInvocation->inExprContextCB;
-	currentInvocation->inExprContextCB = true;
-	_closeIteration(ctxData);
-	currentInvocation->inExprContextCB = saveInExprCtxCB;
+	/*
+	 * Even if there is an invocation already on the stack, there is no
+	 * convincing reason to think this callback belongs to it; PostgreSQL
+	 * will make this callback when the expression context we did belong to
+	 * is being torn down. This is not a hot operation; it only happens in
+	 * rare cases when an SRF has been called and not completely consumed.
+	 * So just unconditionally set up a context for this call, and clean up
+	 * our own mess.
+	 */
+	PG_TRY();
+	{
+		Invocation_pushInvocation(&ctx);
+		currentInvocation->inExprContextCB = true;
+		_closeIteration(ctxData);
+		Invocation_popInvocation(false);
+	}
+	PG_CATCH();
+	{
+		Invocation_popInvocation(true);
+		PG_RE_THROW();
+	}
+	PG_END_TRY();
 }
 
 static Type _getCoerce(Type self, Type other, Oid fromOid, Oid toOid,
@@ -248,6 +271,7 @@ static Type _getCoerce(Type self, Type other, Oid fromOid, Oid toOid,
 	case COERCION_PATH_NONE:
 		elog(ERROR, "no conversion function from (regtype) %d to %d",
 			 fromOid, toOid);
+		pg_unreachable(); /*elog(ERROR is already so marked; what's with gcc?*/
 	case COERCION_PATH_RELABELTYPE:
 		/*
 		 * Binary compatible type. No need for a special coercer.
@@ -260,9 +284,11 @@ static Type _getCoerce(Type self, Type other, Oid fromOid, Oid toOid,
 	case COERCION_PATH_COERCEVIAIO:
 		elog(ERROR, "COERCEVIAIO not implemented from (regtype) %d to %d",
 			 fromOid, toOid);
+		pg_unreachable();
 	case COERCION_PATH_ARRAYCOERCE:
 		elog(ERROR, "ARRAYCOERCE not implemented from (regtype) %d to %d",
 			 fromOid, toOid);
+		pg_unreachable();
 	case COERCION_PATH_FUNC:
 		break;
 	}
@@ -411,11 +437,6 @@ const char* Type_getJNISignature(Type self)
 	return self->typeClass->getJNISignature(self);
 }
 
-const char* Type_getJNIReturnSignature(Type self, bool forMultiCall, bool useAltRepr)
-{
-	return self->typeClass->getJNIReturnSignature(self, forMultiCall, useAltRepr);
-}
-
 Type Type_getArrayType(Type self, Oid arrayTypeId)
 {
 	Type arrayType = self->arrayType;
@@ -460,14 +481,14 @@ TupleDesc Type_getTupleDesc(Type self, PG_FUNCTION_ARGS)
 	return self->typeClass->getTupleDesc(self, fcinfo);
 }
 
-Datum Type_invoke(Type self, jclass cls, jmethodID method, jvalue* args, PG_FUNCTION_ARGS)
+Datum Type_invoke(Type self, Function fn, PG_FUNCTION_ARGS)
 {
-	return self->typeClass->invoke(self, cls, method, args, fcinfo);
+	return self->typeClass->invoke(self, fn, fcinfo);
 }
 
-Datum Type_invokeSRF(Type self, jclass cls, jmethodID method, jvalue* args, PG_FUNCTION_ARGS)
+Datum Type_invokeSRF(Type self, Function fn, PG_FUNCTION_ARGS)
 {
-	bool hasRow;
+	jobject row;
 	CallContextData* ctxData;
 	FuncCallContext* context;
 	MemoryContext currCtx;
@@ -494,10 +515,10 @@ Datum Type_invokeSRF(Type self, jclass cls, jmethodID method, jvalue* args, PG_F
 		 */
 		currCtx = MemoryContextSwitchTo(context->multi_call_memory_ctx);
 
-		/* Call the declared Java function. It returns an instance that can produce
-		 * the rows.
+		/* Call the declared Java function. It returns an instance
+		 * that can produce the rows.
 		 */
-		tmp = Type_getSRFProducer(self, cls, method, args);
+		tmp = pljava_Function_refInvoke(fn);
 		if(tmp == 0)
 		{
 			Invocation_assertDisconnect();
@@ -510,6 +531,7 @@ Datum Type_invokeSRF(Type self, jclass cls, jmethodID method, jvalue* args, PG_F
 		context->user_fctx = ctxData;
 
 		ctxData->elemType = self;
+		ctxData->fn = fn;
 		ctxData->rowProducer = JNI_newGlobalRef(tmp);
 		JNI_deleteLocalRef(tmp);
 
@@ -523,7 +545,6 @@ Datum Type_invokeSRF(Type self, jclass cls, jmethodID method, jvalue* args, PG_F
 			JNI_deleteLocalRef(tmp);
 		}		
 
-		ctxData->trusted    = currentInvocation->trusted;
 		stashCallContext(ctxData);
 
 		/* Register callback to be called when the function ends
@@ -553,12 +574,12 @@ Datum Type_invokeSRF(Type self, jclass cls, jmethodID method, jvalue* args, PG_F
 	currentInvocation->hasConnected = ctxData->hasConnected;
 	currentInvocation->invocation   = ctxData->invocation;
 
-	hasRow = Type_hasNextSRF(self, ctxData->rowProducer, ctxData->rowCollector,
-		(jlong)context->call_cntr);
-
-	if(hasRow)
+	if(JNI_TRUE == pljava_Function_vpcInvoke(ctxData->fn,
+		ctxData->rowProducer, ctxData->rowCollector, (jlong)context->call_cntr,
+		JNI_FALSE, &row))
 	{
-		Datum result = Type_nextSRF(self, ctxData->rowProducer, ctxData->rowCollector);
+		Datum result = Type_datumFromSRF(self, row, ctxData->rowCollector);
+		JNI_deleteLocalRef(row);
 		stashCallContext(ctxData);
 		currentInvocation->hasConnected = false;
 		currentInvocation->invocation   = 0;
@@ -646,6 +667,82 @@ Type Type_fromOidCache(Oid typeId)
 	return (Type)HashMap_getByOid(s_typeByOid, typeId);
 }
 
+/*
+ * Return NULL unless typeId represents a MappedUDT as found in the typeMap,
+ * in which case return a freshly-registered UDT Type.
+ *
+ * A MappedUDT's supporting functions don't have SQL declarations, from which
+ * an ordinary function's PLPrincipal and initiating class loader would be
+ * determined, so when obtaining the support function handles below, NULL will
+ * be passed as the language name, indicating that information isn't available,
+ * and won't be baked into the handles.
+ *
+ * A MappedUDT only has the two support functions readSQL and writeSQL.
+ * The I/O support functions parse and toString are only for a BaseUDT, so
+ * they do not need to be looked up here.
+ *
+ * The typeStruct argument supplies the type's name and namespace to
+ * UDT_registerUDT, as well as the by-value, length, and alignment common to
+ * any registered Type.
+ *
+ * A complication, though: in principle, this is a function on two variables,
+ * typeId and typeMap. (The typeStruct is functionally dependent on typeId.)
+ * But registration of the first one to be encountered will enter it in caches
+ * that depend only on the typeId (or Java class name, for the other direction)
+ * from that point on. This is longstanding PL/Java behavior, but XXX.
+ */
+static inline Type
+checkTypeMappedUDT(Oid typeId, jobject typeMap, Form_pg_type typeStruct)
+{
+	jobject joid;
+	jclass  typeClass;
+	Type    type;
+	jobject readMH;
+	jobject writeMH;
+	TupleDesc tupleDesc;
+	bool    hasTupleDesc;
+
+	if ( NULL == typeMap )
+		return NULL;
+
+	joid      = Oid_create(typeId);
+	typeClass = (jclass)JNI_callObjectMethod(typeMap, s_Map_get, joid);
+	JNI_deleteLocalRef(joid);
+
+	if ( NULL == typeClass )
+		return NULL;
+
+	if ( -2 == typeStruct->typlen )
+	{
+		JNI_deleteLocalRef(typeClass);
+		ereport(ERROR, (
+			errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+			errmsg(
+				"type mapping in PL/Java for %s with NUL-terminated(-2) "
+				"storage not supported",
+				format_type_be_qualified(typeId))
+		));
+	}
+
+	readMH  = pljava_Function_udtReadHandle( typeClass, NULL, true);
+	writeMH = pljava_Function_udtWriteHandle(typeClass, NULL, true);
+
+	tupleDesc = lookup_rowtype_tupdesc_noerror(typeId, -1, true);
+	hasTupleDesc = NULL != tupleDesc;
+	if ( hasTupleDesc )
+		ReleaseTupleDesc(tupleDesc);
+
+	type = (Type)UDT_registerUDT(
+		typeClass, typeId, typeStruct, hasTupleDesc, false,
+		NULL, readMH, writeMH, NULL);
+	/*
+	 * UDT_registerUDT calls JNI_deleteLocalRef on readMH and writeMH.
+	 */
+
+	JNI_deleteLocalRef(typeClass);
+	return type;
+}
+
 Type Type_fromOid(Oid typeId, jobject typeMap)
 {
 	CacheEntry   ce;
@@ -653,7 +750,7 @@ Type Type_fromOid(Oid typeId, jobject typeMap)
 	Form_pg_type typeStruct;
 	Type         type = Type_fromOidCache(typeId);
 
-	if(type != 0)
+	if ( NULL != type )
 		return type;
 
 	typeTup    = PgObject_getValidTuple(TYPEOID, typeId, "type");
@@ -661,12 +758,14 @@ Type Type_fromOid(Oid typeId, jobject typeMap)
 
 	if(typeStruct->typelem != 0 && typeStruct->typlen == -1)
 	{
-		type = Type_getArrayType(Type_fromOid(typeStruct->typelem, typeMap), typeId);
+		type = Type_getArrayType(
+			Type_fromOid(typeStruct->typelem, typeMap), typeId);
 		goto finally;
 	}
 
 	/* For some reason, the anyarray is *not* an array with anyelement as the
 	 * element type. We'd like to see it that way though.
+	 * XXX would we, or does that mistake something intended in PostgreSQL?
 	 */
 	if(typeId == ANYARRAYOID)
 	{
@@ -683,40 +782,35 @@ Type Type_fromOid(Oid typeId, jobject typeMap)
 		goto finally;
 	}
 
-	if(typeMap != 0)
-	{
-		jobject joid      = Oid_create(typeId);
-		jclass  typeClass = (jclass)JNI_callObjectMethod(typeMap, s_Map_get, joid);
-
-		JNI_deleteLocalRef(joid);
-		if(typeClass != 0)
-		{
-			TupleDesc tupleDesc = lookup_rowtype_tupdesc_noerror(typeId, -1, true);
-			bool hasTupleDesc = NULL != tupleDesc;
-			if ( hasTupleDesc )
-				ReleaseTupleDesc(tupleDesc);
-			type = (Type)UDT_registerUDT(typeClass, typeId, typeStruct, hasTupleDesc, false);
-			JNI_deleteLocalRef(typeClass);
-			goto finally;
-		}
-	}
+	/*
+	 * Perhaps we have found a MappedUDT. If so, this check will register and
+	 * return it.
+	 */
+	type = checkTypeMappedUDT(typeId, typeMap, typeStruct);
+	if ( NULL != type )
+		goto finally;
 
 	/* Composite and record types will not have a TypeObtainer registered
 	 */
-	if(typeStruct->typtype == 'c' || (typeStruct->typtype == 'p' && typeId == RECORDOID))
+	if(typeStruct->typtype == 'c'
+		|| (typeStruct->typtype == 'p' && typeId == RECORDOID))
 	{
 		type = Composite_obtain(typeId);
 		goto finally;
 	}
 
 	ce = (CacheEntry)HashMap_getByOid(s_obtainerByOid, typeId);
-	if(ce == 0)
+	if ( NULL == ce )
 	{
-		type = Function_checkTypeUDT(typeId, typeStruct);
-		if ( 0 != type )
+		/*
+		 * Perhaps we have found a BaseUDT. If so, this check will register and
+		 * return it.
+		 */
+		type = Function_checkTypeBaseUDT(typeId, typeStruct);
+		if ( NULL != type )
 			goto finally;
 		/*
-		 * Default to String and standard textin/textout coersion.
+		 * Default to String and standard textin/textout coercion.
 		 * Note: if the AS spec includes a Java signature, and the corresponding
 		 * Java type is not String, that will trigger a call to
 		 * Type_fromJavaType to see if a mapping is registered that way. If not,
@@ -749,11 +843,11 @@ bool _Type_canReplaceType(Type self, Type other)
 	return self->typeClass == other->typeClass;
 }
 
-Datum _Type_invoke(Type self, jclass cls, jmethodID method, jvalue* args, PG_FUNCTION_ARGS)
+Datum _Type_invoke(Type self, Function fn, PG_FUNCTION_ARGS)
 {
 	MemoryContext currCtx;
 	Datum ret;
-	jobject value = JNI_callStaticObjectMethodA(cls, method, args);
+	jobject value = pljava_Function_refInvoke(fn);
 	if(value == 0)
 	{
 		fcinfo->isnull = true;
@@ -775,36 +869,14 @@ static Type _Type_createArrayType(Type self, Oid arrayTypeId)
 	return Array_fromOid(arrayTypeId, self);
 }
 
-static jobject _Type_getSRFProducer(Type self, jclass cls, jmethodID method, jvalue* args)
-{
-	return JNI_callStaticObjectMethodA(cls, method, args);
-}
-
 static jobject _Type_getSRFCollector(Type self, PG_FUNCTION_ARGS)
 {
 	return 0;
 }
 
-static bool _Type_hasNextSRF(Type self, jobject rowProducer, jobject rowCollector, jlong callCounter)
+static Datum _Type_datumFromSRF(Type self, jobject row, jobject rowCollector)
 {
-	return (JNI_callBooleanMethod(rowProducer, s_Iterator_hasNext) == JNI_TRUE);
-}
-
-static Datum _Type_nextSRF(Type self, jobject rowProducer, jobject rowCollector)
-{
-	jobject tmp = JNI_callObjectMethod(rowProducer, s_Iterator_next);
-	Datum result = Type_coerceObject(self, tmp);
-	JNI_deleteLocalRef(tmp);
-	return result;
-}
-
-static void _Type_closeSRF(Type self, jobject rowProducer)
-{
-}
-
-jobject Type_getSRFProducer(Type self, jclass cls, jmethodID method, jvalue* args)
-{
-	return self->typeClass->getSRFProducer(self, cls, method, args);
+	return Type_coerceObject(self, row);
 }
 
 jobject Type_getSRFCollector(Type self, PG_FUNCTION_ARGS)
@@ -812,19 +884,9 @@ jobject Type_getSRFCollector(Type self, PG_FUNCTION_ARGS)
 	return self->typeClass->getSRFCollector(self, fcinfo);
 }
 
-bool Type_hasNextSRF(Type self, jobject rowProducer, jobject rowCollector, jlong callCounter)
+Datum Type_datumFromSRF(Type self, jobject row, jobject rowCollector)
 {
-	return self->typeClass->hasNextSRF(self, rowProducer, rowCollector, callCounter);
-}
-
-Datum Type_nextSRF(Type self, jobject rowProducer, jobject rowCollector)
-{
-	return self->typeClass->nextSRF(self, rowProducer, rowCollector);
-}
-
-void Type_closeSRF(Type self, jobject rowProducer)
-{
-	self->typeClass->closeSRF(self, rowProducer);
+	return self->typeClass->datumFromSRF(self, row, rowCollector);
 }
 
 static Type _Type_getRealType(Type self, Oid realId, jobject typeMap)
@@ -837,11 +899,6 @@ static const char* _Type_getJNISignature(Type self)
 	return self->typeClass->JNISignature;
 }
 
-static const char* _Type_getJNIReturnSignature(Type self, bool forMultiCall, bool useAltRepr)
-{
-	return forMultiCall ? "Ljava/util/Iterator;" : Type_getJNISignature(self);
-}
-
 TupleDesc _Type_getTupleDesc(Type self, PG_FUNCTION_ARGS)
 {
 	ereport(ERROR,
@@ -985,11 +1042,15 @@ void Type_initialize(void)
 	pljava_SQLXMLImpl_initialize();
 
 	s_Map_class = JNI_newGlobalRef(PgObject_getJavaClass("java/util/Map"));
-	s_Map_get = PgObject_getJavaMethod(s_Map_class, "get", "(Ljava/lang/Object;)Ljava/lang/Object;");
+	s_Map_get = PgObject_getJavaMethod(
+		s_Map_class, "get", "(Ljava/lang/Object;)Ljava/lang/Object;");
 
-	s_Iterator_class = JNI_newGlobalRef(PgObject_getJavaClass("java/util/Iterator"));
-	s_Iterator_hasNext = PgObject_getJavaMethod(s_Iterator_class, "hasNext", "()Z");
-	s_Iterator_next = PgObject_getJavaMethod(s_Iterator_class, "next", "()Ljava/lang/Object;");
+	s_Iterator_class = JNI_newGlobalRef(
+		PgObject_getJavaClass("java/util/Iterator"));
+	s_Iterator_hasNext = PgObject_getJavaMethod(
+		s_Iterator_class, "hasNext", "()Z");
+	s_Iterator_next = PgObject_getJavaMethod(
+		s_Iterator_class, "next", "()Ljava/lang/Object;");
 
 #if PG_VERSION_NUM < 110000
 	BOOLARRAYOID   = get_array_type(BOOLOID);
@@ -1009,10 +1070,12 @@ void Type_initialize(void)
  */
 TypeClass TypeClass_alloc(const char* typeName)
 {
-	return TypeClass_alloc2(typeName, sizeof(struct TypeClass_), sizeof(struct Type_));
+	return TypeClass_alloc2(
+		typeName, sizeof(struct TypeClass_), sizeof(struct Type_));
 }
 
-TypeClass TypeClass_alloc2(const char* typeName, Size classSize, Size instanceSize)
+TypeClass TypeClass_alloc2(
+	const char* typeName, Size classSize, Size instanceSize)
 {
 	TypeClass self = (TypeClass)MemoryContextAlloc(TopMemoryContext, classSize);
 	PgObjectClass_init((PgObjectClass)self, typeName, instanceSize, 0);
@@ -1024,14 +1087,10 @@ TypeClass TypeClass_alloc2(const char* typeName, Size classSize, Size instanceSi
 	self->coerceObject    = (ObjectCoercer)_PgObject_pureVirtualCalled;
 	self->createArrayType = _Type_createArrayType;
 	self->invoke          = _Type_invoke;
-	self->getSRFProducer  = _Type_getSRFProducer;
 	self->getSRFCollector = _Type_getSRFCollector;
-	self->hasNextSRF      = _Type_hasNextSRF;
-	self->nextSRF         = _Type_nextSRF;
-	self->closeSRF        = _Type_closeSRF;
+	self->datumFromSRF    = _Type_datumFromSRF;
 	self->getTupleDesc    = _Type_getTupleDesc;
 	self->getJNISignature = _Type_getJNISignature;
-	self->getJNIReturnSignature = _Type_getJNIReturnSignature;
 	self->dynamic         = false;
 	self->outParameter    = false;
 	self->getRealType     = _Type_getRealType;
@@ -1051,7 +1110,8 @@ Type TypeClass_allocInstance(TypeClass cls, Oid typeId)
  */
 Type TypeClass_allocInstance2(TypeClass cls, Oid typeId, Form_pg_type pgType)
 {
-	Type t = (Type)PgObjectClass_allocInstance((PgObjectClass)(cls), TopMemoryContext);
+	Type t = (Type)
+		PgObjectClass_allocInstance((PgObjectClass)(cls), TopMemoryContext);
 	t->typeId       = typeId;
 	t->arrayType    = 0;
 	t->elementType  = 0;
@@ -1083,9 +1143,11 @@ Type TypeClass_allocInstance2(TypeClass cls, Oid typeId, Form_pg_type pgType)
 /*
  * Register this type.
  */
-static void _registerType(Oid typeId, const char* javaTypeName, Type type, TypeObtainer obtainer)
+static void _registerType(
+	Oid typeId, const char* javaTypeName, Type type, TypeObtainer obtainer)
 {
-	CacheEntry ce = (CacheEntry)MemoryContextAlloc(TopMemoryContext, sizeof(CacheEntryData));
+	CacheEntry ce = (CacheEntry)
+		MemoryContextAlloc(TopMemoryContext, sizeof(CacheEntryData));
 	ce->typeId   = typeId;
 	ce->type     = type;
 	ce->obtainer = obtainer;
@@ -1111,10 +1173,13 @@ static void _registerType(Oid typeId, const char* javaTypeName, Type type, TypeO
 
 void Type_registerType(const char* javaTypeName, Type type)
 {
-	_registerType(type->typeId, javaTypeName, type, (TypeObtainer)_PgObject_pureVirtualCalled);
+	_registerType(
+		type->typeId, javaTypeName, type,
+		(TypeObtainer)_PgObject_pureVirtualCalled);
 }
 
-void Type_registerType2(Oid typeId, const char* javaTypeName, TypeObtainer obtainer)
+void Type_registerType2(
+	Oid typeId, const char* javaTypeName, TypeObtainer obtainer)
 {
 	_registerType(typeId, javaTypeName, 0, obtainer);
 }
