@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004-2019 Tada AB and other contributors, as listed below.
+ * Copyright (c) 2004-2021 Tada AB and other contributors, as listed below.
  *
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the The BSD 3-Clause License
@@ -82,8 +82,14 @@ extern int vsnprintf(char* buf, size_t count, const char* format, va_list arg);
 #define PG_INT32_MAX    (0x7FFFFFFF)
 #endif
 
+/*
+ * This symbol was spelled without the underscores prior to PG 14.
+ */
+#if PG_VERSION_NUM < 140000
+#define PG_NODE_TREEOID PGNODETREEOID
+#endif
+
 extern void* mainThreadId;
-extern bool pljavaEntryFence(JNIEnv* env);
 extern JNIEnv* currentJNIEnv;
 extern MemoryContext JavaMemoryContext;
 
@@ -183,14 +189,6 @@ typedef union
 {
 	void*  ptrVal;
 	jlong  longVal; /* 64 bit quantity */
-	struct
-	{
-		/* Used when calculating pointer hash in systems where
-		 * a pointer is 64 bit
-		 */
-		uint32 intVal_1;
-		uint32 intVal_2;
-	} x64;
 } Ptr2Long;
 
 struct Invocation_;
