@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004-2019 Tada AB and other contributors, as listed below.
+ * Copyright (c) 2004-2020 Tada AB and other contributors, as listed below.
  *
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the The BSD 3-Clause License
@@ -24,8 +24,13 @@ import org.postgresql.pljava.ObjectPool;
 import org.postgresql.pljava.PooledObject;
 import org.postgresql.pljava.SavepointListener;
 import org.postgresql.pljava.TransactionListener;
+import org.postgresql.pljava.sqlgen.Lexicals.Identifier;
+
 import org.postgresql.pljava.jdbc.SQLUtils;
 
+import org.postgresql.pljava.elog.ELogHandler;
+
+import static org.postgresql.pljava.internal.Backend.doInPG;
 
 /**
  * An instance of this interface reflects the current session. The attribute
@@ -39,6 +44,7 @@ import org.postgresql.pljava.jdbc.SQLUtils;
  */
 public class Session implements org.postgresql.pljava.Session
 {
+	@SuppressWarnings("removal")
 	private final TransactionalMap m_attributes = new TransactionalMap(new HashMap());
 
 	/**
@@ -94,6 +100,8 @@ public class Session implements org.postgresql.pljava.Session
 	 * Java and kept in sync by using a {@link TransactionListener}.
 	 */
 	@Override
+	@SuppressWarnings("removal")
+	@Deprecated(since="1.5.3", forRemoval=true)
 	public Object getAttribute(String attributeName)
 	{
 		return m_attributes.get(attributeName);
@@ -117,6 +125,8 @@ public class Session implements org.postgresql.pljava.Session
 	}
 
 	@Override
+	@SuppressWarnings("removal")
+	@Deprecated(since="1.5.0", forRemoval=true)
 	public String getSessionUserName()
 	{
 		return getOuterUserName();
@@ -133,6 +143,8 @@ public class Session implements org.postgresql.pljava.Session
 	 * Java and kept in sync by using a {@link TransactionListener}.
 	 */
 	@Override
+	@SuppressWarnings("removal")
+	@Deprecated(since="1.5.3", forRemoval=true)
 	public void removeAttribute(String attributeName)
 	{
 		m_attributes.remove(attributeName);
@@ -148,6 +160,8 @@ public class Session implements org.postgresql.pljava.Session
 	 * Java and kept in sync by using a {@link TransactionListener}.
 	 */
 	@Override
+	@SuppressWarnings("removal")
+	@Deprecated(since="1.5.3", forRemoval=true)
 	public void setAttribute(String attributeName, Object value)
 	{
 		m_attributes.put(attributeName, value);
@@ -172,6 +186,8 @@ public class Session implements org.postgresql.pljava.Session
 	}
 
 	@Override
+	@SuppressWarnings("removal")
+	@Deprecated(since="1.5.0", forRemoval=true)
 	public void executeAsSessionUser(Connection conn, String statement)
 	throws SQLException
 	{
@@ -183,7 +199,7 @@ public class Session implements org.postgresql.pljava.Session
 	throws SQLException
 	{
 		Statement stmt = conn.createStatement();
-		synchronized(Backend.THREADLOCK)
+		doInPG(() ->
 		{
 			ResultSet rs = null;
 			AclId outerUser = AclId.getOuterUser();
@@ -207,7 +223,7 @@ public class Session implements org.postgresql.pljava.Session
 				if ( changeSucceeded )
 					_setUser(effectiveUser, wasLocalChange);
 			}
-		}
+		});
 	}
 
 	/**
@@ -215,24 +231,24 @@ public class Session implements org.postgresql.pljava.Session
 	 * Currently used only in Commands.java. Not made visible API yet
 	 * because there <em>has</em> to be a more general way to do this.
 	 */
-	public String getOuterUserSchema()
+	public Identifier.Simple getOuterUserSchema()
 	throws SQLException
 	{
 		Statement stmt = SQLUtils.getDefaultConnection().createStatement();
-		synchronized(Backend.THREADLOCK)
+		return doInPG(() ->
 		{
 			ResultSet rs = null;
-			AclId sessionUser = AclId.getSessionUser();
+			AclId outerUser = AclId.getOuterUser();
 			AclId effectiveUser = AclId.getUser();
 			boolean wasLocalChange = false;
 			boolean changeSucceeded = false;
 			try
 			{
-				wasLocalChange = _setUser(sessionUser, true);
+				wasLocalChange = _setUser(outerUser, true);
 				changeSucceeded = true;
 				rs = stmt.executeQuery("SELECT current_schema()");
 				if ( rs.next() )
-					return rs.getString(1);
+					return Identifier.Simple.fromCatalog(rs.getString(1));
 				throw new SQLException("Unable to obtain current schema");
 			}
 			finally
@@ -242,7 +258,7 @@ public class Session implements org.postgresql.pljava.Session
 				if ( changeSucceeded )
 					_setUser(effectiveUser, wasLocalChange);
 			}
-		}
+		});
 	}
 
 	/**
