@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2019 Tada AB and other contributors, as listed below.
+ * Copyright (c) 2019-2020 Tada AB and other contributors, as listed below.
  *
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the The BSD 3-Clause License
@@ -27,6 +27,7 @@ import java.util.concurrent.Callable;
 import java.util.concurrent.CountDownLatch;
 import java.util.concurrent.ExecutorService;
 import java.util.concurrent.Executors;
+import static java.util.concurrent.Executors.privilegedCallable;
 import java.util.concurrent.Future;
 import java.util.concurrent.LinkedBlockingQueue;
 import java.util.concurrent.ThreadFactory;
@@ -34,6 +35,8 @@ import java.util.concurrent.ThreadFactory;
 import java.util.concurrent.CancellationException;
 import java.util.concurrent.ExecutionException;
 
+import static org.postgresql.pljava.internal.Backend.doInPG;
+
 /**
  * Interface that wraps a PostgreSQL native variable-length ("varlena") datum;
  * implementing classes present an existing one to Java as a readable
@@ -285,13 +288,9 @@ public interface VarlenaWrapper extends Closeable
 								"Verifier finished prematurely");
 					}
 				}
-				catch ( SQLException sqe )
+				catch ( SQLException | RuntimeException e )
 				{
-					throw sqe;
-				}
-				catch ( RuntimeException rte )
-				{
-					throw rte;
+					throw e;
 				}
 				catch ( Exception e )
 				{
@@ -368,13 +367,13 @@ public interface VarlenaWrapper extends Closeable
 				{
 					if ( null != m_buf )
 						return m_buf;
-					synchronized ( Backend.THREADLOCK )
+					doInPG(() ->
 					{
 						m_buf = _detoast(
 							m_varlena, guardedLong(), m_snapshot,
 							m_resourceOwner).asReadOnlyBuffer();
 						m_snapshot = 0;
-					}
+					});
 					return m_buf;
 				}
 				finally
@@ -707,7 +706,7 @@ public interface VarlenaWrapper extends Closeable
 					if ( 0 < m_buf.remaining()  &&  0 < desiredCapacity )
 						return m_buf;
 					ByteBuffer filledBuf = m_buf;
-					synchronized ( Backend.THREADLOCK )
+					doInPG(() ->
 					{
 						int lstate = lock(true); // true -> upgrade my held pin
 						try
@@ -719,7 +718,7 @@ public interface VarlenaWrapper extends Closeable
 						{
 							unlock(lstate);
 						}
-					}
+					});
 					m_verifier.update(this, filledBuf);
 					if ( 0 == desiredCapacity )
 						m_verifier.update(MarkableSequenceInputStream.NO_MORE);
@@ -940,18 +939,11 @@ public interface VarlenaWrapper extends Closeable
 		@Override
 		public final Void call() throws Exception
 		{
-			InputStream is = null;
-			try
+			try ( InputStream is = new MarkableSequenceInputStream(m_queue) )
 			{
-				is = new MarkableSequenceInputStream(m_queue);
 				verify(is);
+				return null;
 			}
-			finally
-			{
-				if ( null != is )
-					is.close();
-			}
-			return null;
 		}
 
 		/**
@@ -1039,7 +1031,9 @@ public interface VarlenaWrapper extends Closeable
 			{
 				if ( 1 == m_latch.getCount() )
 				{
-					m_future = LazyExecutorService.INSTANCE.submit(this);
+					m_future =
+						LazyExecutorService.INSTANCE
+							.submit(privilegedCallable(this));
 					m_latch.countDown();
 				}
 			}
